#!/usr/bin/env python3
"""
# nvmept_pi.py
#
# Test fio's io_uring_cmd ioengine support for DIF/DIX end-to-end data
# protection.
#
# USAGE
# see python3 nvmept_pi.py --help
#
# EXAMPLES (THIS IS A DESTRUCTIVE TEST!!)
# python3 t/nvmept_pi.py --dut /dev/ng0n1 -f ./fio
# python3 t/nvmept_pi.py --dut /dev/ng0n1 -f ./fio --lbaf 1
#
# REQUIREMENTS
# Python 3.6
#
"""
import os
import sys
import json
import time
import locale
import logging
import argparse
import itertools
import subprocess
from pathlib import Path
from fiotestlib import FioJobCmdTest, run_fio_tests
from fiotestcommon import SUCCESS_NONZERO

NUMBER_IOS = 8192
BS_LOW = 1
BS_HIGH = 16

class DifDixTest(FioJobCmdTest):
    """
    NVMe DIF/DIX test class.
    """

    def setup(self, parameters):
        """Setup a test."""

        fio_args = [
            "--name=nvmept_pi",
            f"--ioengine={self.fio_opts['ioengine']}",
            f"--filename={self.fio_opts['filename']}",
            f"--rw={self.fio_opts['rw']}",
            f"--bsrange={self.fio_opts['bsrange']}",
            f"--output={self.filenames['output']}",
            f"--md_per_io_size={self.fio_opts['md_per_io_size']}",
            f"--pi_act={self.fio_opts['pi_act']}",
            f"--pi_chk={self.fio_opts['pi_chk']}",
            f"--apptag={self.fio_opts['apptag']}",
            f"--apptag_mask={self.fio_opts['apptag_mask']}",
        ]
        for opt in ['fixedbufs', 'nonvectored', 'force_async', 'registerfiles',
                    'sqthread_poll', 'sqthread_poll_cpu', 'hipri', 'nowait',
                    'time_based', 'runtime', 'verify', 'io_size', 'offset', 'number_ios',
                    'output-format']:
            if opt in self.fio_opts:
                option = f"--{opt}={self.fio_opts[opt]}"
                fio_args.append(option)

        if self.fio_opts['ioengine'] == 'io_uring_cmd':
            fio_args.append('--cmd_type=nvme')
        elif self.fio_opts['ioengine'] == 'xnvme':
            fio_args.append('--thread=1')
            fio_args.append('--xnvme_async=io_uring_cmd')

        super().setup(fio_args)


TEST_LIST = [
#
# Write data with pi_act=1 and then read the data back (with both
# pi_act=[0,1]).
#
    {
        # Write workload with variable IO sizes
        # pi_act=1
        "test_id": 101,
        "fio_opts": {
            "rw": 'write',
            "number_ios": NUMBER_IOS,
            "output-format": "json",
            "apptag": "0x8888",
            "apptag_mask": "0xFFFF",
            "pi_act": 1,
            },
        "pi_chk": "APPTAG,GUARD,REFTAG",
        "bs_low": BS_LOW,
        "bs_high": BS_HIGH,
        "test_class": DifDixTest,
    },
    {
        # Read workload with fixed small IO size
        # pi_act=0
        "test_id": 102,
        "fio_opts": {
            "rw": 'read',
            "number_ios": NUMBER_IOS,
            "output-format": "json",
            "pi_act": 0,
            "apptag": "0x8888",
            "apptag_mask": "0xFFFF",
            },
        "pi_chk": "APPTAG,GUARD,REFTAG",
        "bs_low": BS_LOW,
        "bs_high": BS_LOW,
        "test_class": DifDixTest,
    },
    {
        # Read workload with fixed small IO size
        # pi_act=1
        "test_id": 103,
        "fio_opts": {
            "rw": 'read',
            "number_ios": NUMBER_IOS,
            "output-format": "json",
            "pi_act": 1,
            "apptag": "0x8888",
            "apptag_mask": "0xFFFF",
            },
        "pi_chk": "APPTAG,GUARD,REFTAG",
        "bs_low": BS_LOW,
        "bs_high": BS_LOW,
        "test_class": DifDixTest,
    },
    {
        # Write workload with fixed large IO size
        # Precondition for read workloads to follow
        # pi_act=1
        "test_id": 104,
        "fio_opts": {
            "rw": 'write',
            "number_ios": NUMBER_IOS,
            "output-format": "json",
            "apptag": "0x8888",
            "apptag_mask": "0xFFFF",
            "pi_act": 1,
            },
        "pi_chk": "APPTAG,GUARD,REFTAG",
        "bs_low": BS_HIGH,
        "bs_high": BS_HIGH,
        "test_class": DifDixTest,
    },
    {
        # Read workload with variable IO sizes
        # pi_act=0
        "test_id": 105,
        "fio_opts": {
            "rw": 'read',
            "number_ios": NUMBER_IOS,
            "output-format": "json",
            "pi_act": 0,
            "apptag": "0x8888",
            "apptag_mask": "0xFFFF",
            },
        "pi_chk": "APPTAG,GUARD,REFTAG",
        "bs_low": BS_LOW,
        "bs_high": BS_HIGH,
        "test_class": DifDixTest,
    },
    {
        # Read workload with variable IO sizes
        # pi_act=1
        "test_id": 106,
        "fio_opts": {
            "rw": 'read',
            "number_ios": NUMBER_IOS,
            "output-format": "json",
            "pi_act": 1,
            "apptag": "0x8888",
            "apptag_mask": "0xFFFF",
            },
        "pi_chk": "APPTAG,GUARD,REFTAG",
        "bs_low": BS_LOW,
        "bs_high": BS_HIGH,
        "test_class": DifDixTest,
    },
#
# Write data with pi_act=0 and then read the data back (with both
# pi_act=[0,1]).
#
    {
        # Write workload with variable IO sizes
        # pi_act=0
        "test_id": 201,
        "fio_opts": {
            "rw": 'write',
            "number_ios": NUMBER_IOS,
            "output-format": "json",
            "apptag": "0x8888",
            "apptag_mask": "0xFFFF",
            "pi_act": 0,
            },
        "pi_chk": "APPTAG,GUARD,REFTAG",
        "bs_low": BS_LOW,
        "bs_high": BS_HIGH,
        "test_class": DifDixTest,
    },
    {
        # Read workload with fixed small IO size
        # pi_act=0
        "test_id": 202,
        "fio_opts": {
            "rw": 'read',
            "number_ios": NUMBER_IOS,
            "output-format": "json",
            "pi_act": 0,
            "apptag": "0x8888",
            "apptag_mask": "0xFFFF",
            },
        "pi_chk": "APPTAG,GUARD,REFTAG",
        "bs_low": BS_LOW,
        "bs_high": BS_LOW,
        "test_class": DifDixTest,
    },
    {
        # Read workload with fixed small IO size
        # pi_act=1
        "test_id": 203,
        "fio_opts": {
            "rw": 'read',
            "number_ios": NUMBER_IOS,
            "output-format": "json",
            "pi_act": 1,
            "apptag": "0x8888",
            "apptag_mask": "0xFFFF",
            },
        "pi_chk": "APPTAG,GUARD,REFTAG",
        "bs_low": BS_LOW,
        "bs_high": BS_LOW,
        "test_class": DifDixTest,
    },
    {
        # Write workload with fixed large IO sizes
        # pi_act=0
        "test_id": 204,
        "fio_opts": {
            "rw": 'write',
            "number_ios": NUMBER_IOS,
            "output-format": "json",
            "apptag": "0x8888",
            "apptag_mask": "0xFFFF",
            "pi_act": 0,
            },
        "pi_chk": "APPTAG,GUARD,REFTAG",
        "bs_low": BS_HIGH,
        "bs_high": BS_HIGH,
        "test_class": DifDixTest,
    },
    {
        # Read workload with variable IO sizes
        # pi_act=0
        "test_id": 205,
        "fio_opts": {
            "rw": 'read',
            "number_ios": NUMBER_IOS,
            "output-format": "json",
            "pi_act": 0,
            "apptag": "0x8888",
            "apptag_mask": "0xFFFF",
            },
        "pi_chk": "APPTAG,GUARD,REFTAG",
        "bs_low": BS_LOW,
        "bs_high": BS_HIGH,
        "test_class": DifDixTest,
    },
    {
        # Read workload with variable IO sizes
        # pi_act=1
        "test_id": 206,
        "fio_opts": {
            "rw": 'read',
            "number_ios": NUMBER_IOS,
            "output-format": "json",
            "pi_act": 1,
            "apptag": "0x8888",
            "apptag_mask": "0xFFFF",
            },
        "pi_chk": "APPTAG,GUARD,REFTAG",
        "bs_low": BS_LOW,
        "bs_high": BS_HIGH,
        "test_class": DifDixTest,
    },
#
# Test apptag errors.
#
    {
        # Read workload with variable IO sizes
        # pi_act=0
        # trigger an apptag error
        "test_id": 301,
        "fio_opts": {
            "rw": 'read',
            "number_ios": NUMBER_IOS,
            "output-format": "json",
            "pi_act": 0,
            "apptag": "0x0888",
            "apptag_mask": "0xFFFF",
            },
        "pi_chk": "APPTAG,GUARD,REFTAG",
        "bs_low": BS_LOW,
        "bs_high": BS_HIGH,
        "success": SUCCESS_NONZERO,
        "test_class": DifDixTest,
    },
    {
        # Read workload with variable IO sizes
        # pi_act=1
        # trigger an apptag error
        "test_id": 302,
        "fio_opts": {
            "rw": 'read',
            "number_ios": NUMBER_IOS,
            "output-format": "json",
            "pi_act": 1,
            "apptag": "0x0888",
            "apptag_mask": "0xFFFF",
            },
        "pi_chk": "APPTAG,GUARD,REFTAG",
        "bs_low": BS_LOW,
        "bs_high": BS_HIGH,
        "success": SUCCESS_NONZERO,
        "test_class": DifDixTest,
    },
    {
        # Read workload with variable IO sizes
        # pi_act=0
        # trigger an apptag error
        # same as above but with pi_chk=APPTAG only
        "test_id": 303,
        "fio_opts": {
            "rw": 'read',
            "number_ios": NUMBER_IOS,
            "output-format": "json",
            "pi_act": 0,
            "apptag": "0x0888",
            "apptag_mask": "0xFFFF",
            },
        "pi_chk": "APPTAG",
        "bs_low": BS_LOW,
        "bs_high": BS_HIGH,
        "success": SUCCESS_NONZERO,
        "test_class": DifDixTest,
    },
    {
        # Read workload with variable IO sizes
        # pi_act=1
        # trigger an apptag error
        # same as above but with pi_chk=APPTAG only
        "test_id": 304,
        "fio_opts": {
            "rw": 'read',
            "number_ios": NUMBER_IOS,
            "output-format": "json",
            "pi_act": 1,
            "apptag": "0x0888",
            "apptag_mask": "0xFFFF",
            },
        "pi_chk": "APPTAG",
        "bs_low": BS_LOW,
        "bs_high": BS_HIGH,
        "success": SUCCESS_NONZERO,
        "test_class": DifDixTest,
    },
    {
        # Read workload with variable IO sizes
        # pi_act=0
        # this case would trigger an apptag error, but pi_chk says to check
        # only the Guard PI and reftag, so there should be no error
        "test_id": 305,
        "fio_opts": {
            "rw": 'read',
            "number_ios": NUMBER_IOS,
            "output-format": "json",
            "pi_act": 0,
            "apptag": "0x0888",
            "apptag_mask": "0xFFFF",
            },
        "pi_chk": "GUARD,REFTAG",
        "bs_low": BS_LOW,
        "bs_high": BS_HIGH,
        "test_class": DifDixTest,
    },
    {
        # Read workload with variable IO sizes
        # pi_act=1
        # this case would trigger an apptag error, but pi_chk says to check
        # only the Guard PI and reftag, so there should be no error
        "test_id": 306,
        "fio_opts": {
            "rw": 'read',
            "number_ios": NUMBER_IOS,
            "output-format": "json",
            "pi_act": 1,
            "apptag": "0x0888",
            "apptag_mask": "0xFFFF",
            },
        "pi_chk": "GUARD,REFTAG",
        "bs_low": BS_LOW,
        "bs_high": BS_HIGH,
        "test_class": DifDixTest,
    },
    {
        # Read workload with variable IO sizes
        # pi_act=0
        # this case would trigger an apptag error, but pi_chk says to check
        # only the Guard PI, so there should be no error
        "test_id": 307,
        "fio_opts": {
            "rw": 'read',
            "number_ios": NUMBER_IOS,
            "output-format": "json",
            "pi_act": 0,
            "apptag": "0x0888",
            "apptag_mask": "0xFFFF",
            },
        "pi_chk": "GUARD",
        "bs_low": BS_LOW,
        "bs_high": BS_HIGH,
        "test_class": DifDixTest,
    },
    {
        # Read workload with variable IO sizes
        # pi_act=1
        # this case would trigger an apptag error, but pi_chk says to check
        # only the Guard PI, so there should be no error
        "test_id": 308,
        "fio_opts": {
            "rw": 'read',
            "number_ios": NUMBER_IOS,
            "output-format": "json",
            "pi_act": 1,
            "apptag": "0x0888",
            "apptag_mask": "0xFFFF",
            },
        "pi_chk": "GUARD",
        "bs_low": BS_LOW,
        "bs_high": BS_HIGH,
        "test_class": DifDixTest,
    },
    {
        # Read workload with variable IO sizes
        # pi_act=0
        # this case would trigger an apptag error, but pi_chk says to check
        # only the reftag, so there should be no error
        # This case will be skipped when the device is formatted with Type 3 PI
        # since Type 3 PI ignores the reftag
        "test_id": 309,
        "fio_opts": {
            "rw": 'read',
            "number_ios": NUMBER_IOS,
            "output-format": "json",
            "pi_act": 0,
            "apptag": "0x0888",
            "apptag_mask": "0xFFFF",
            },
        "pi_chk": "REFTAG",
        "bs_low": BS_LOW,
        "bs_high": BS_HIGH,
        "skip": "type3",
        "test_class": DifDixTest,
    },
    {
        # Read workload with variable IO sizes
        # pi_act=1
        # this case would trigger an apptag error, but pi_chk says to check
        # only the reftag, so there should be no error
        # This case will be skipped when the device is formatted with Type 3 PI
        # since Type 3 PI ignores the reftag
        "test_id": 310,
        "fio_opts": {
            "rw": 'read',
            "number_ios": NUMBER_IOS,
            "output-format": "json",
            "pi_act": 1,
            "apptag": "0x0888",
            "apptag_mask": "0xFFFF",
            },
        "pi_chk": "REFTAG",
        "bs_low": BS_LOW,
        "bs_high": BS_HIGH,
        "skip": "type3",
        "test_class": DifDixTest,
    },
    {
        # Read workload with variable IO sizes
        # pi_act=0
        # use apptag mask to ignore apptag mismatch
        "test_id": 311,
        "fio_opts": {
            "rw": 'read',
            "number_ios": NUMBER_IOS,
            "output-format": "json",
            "pi_act": 0,
            "apptag": "0x0888",
            "apptag_mask": "0x0FFF",
            },
        "pi_chk": "APPTAG,GUARD,REFTAG",
        "bs_low": BS_LOW,
        "bs_high": BS_HIGH,
        "test_class": DifDixTest,
    },
    {
        # Read workload with variable IO sizes
        # pi_act=1
        # use apptag mask to ignore apptag mismatch
        "test_id": 312,
        "fio_opts": {
            "rw": 'read',
            "number_ios": NUMBER_IOS,
            "output-format": "json",
            "pi_act": 1,
            "apptag": "0x0888",
            "apptag_mask": "0x0FFF",
            },
        "pi_chk": "APPTAG,GUARD,REFTAG",
        "bs_low": BS_LOW,
        "bs_high": BS_HIGH,
        "test_class": DifDixTest,
    },
    {
        # Read workload with variable IO sizes
        # pi_act=0
        # use apptag mask to ignore apptag mismatch
        "test_id": 313,
        "fio_opts": {
            "rw": 'read',
            "number_ios": NUMBER_IOS,
            "output-format": "json",
            "pi_act": 0,
            "apptag": "0xF888",
            "apptag_mask": "0x0FFF",
            },
        "pi_chk": "APPTAG,GUARD,REFTAG",
        "bs_low": BS_LOW,
        "bs_high": BS_HIGH,
        "test_class": DifDixTest,
    },
    {
        # Read workload with variable IO sizes
        # pi_act=1
        # use apptag mask to ignore apptag mismatch
        "test_id": 314,
        "fio_opts": {
            "rw": 'read',
            "number_ios": NUMBER_IOS,
            "output-format": "json",
            "pi_act": 1,
            "apptag": "0xF888",
            "apptag_mask": "0x0FFF",
            },
        "pi_chk": "APPTAG,GUARD,REFTAG",
        "bs_low": BS_LOW,
        "bs_high": BS_HIGH,
        "test_class": DifDixTest,
    },
    {
        # Write workload with fixed large IO sizes
        # Set apptag=0xFFFF to disable all checking for Type 1 and 2
        # pi_act=1
        "test_id": 315,
        "fio_opts": {
            "rw": 'write',
            "number_ios": NUMBER_IOS,
            "output-format": "json",
            "apptag": "0xFFFF",
            "apptag_mask": "0xFFFF",
            "pi_act": 1,
            },
        "pi_chk": "APPTAG,GUARD,REFTAG",
        "bs_low": BS_HIGH,
        "bs_high": BS_HIGH,
        "skip": "type3",
        "test_class": DifDixTest,
    },
    {
        # Read workload with variable IO sizes
        # pi_act=0
        # Data was written with apptag=0xFFFF
        # Reading the data back should disable all checking for Type 1 and 2
        "test_id": 316,
        "fio_opts": {
            "rw": 'read',
            "number_ios": NUMBER_IOS,
            "output-format": "json",
            "pi_act": 0,
            "apptag": "0x0101",
            "apptag_mask": "0xFFFF",
            },
        "pi_chk": "APPTAG,GUARD,REFTAG",
        "bs_low": BS_LOW,
        "bs_high": BS_HIGH,
        "skip": "type3",
        "test_class": DifDixTest,
    },
    {
        # Read workload with variable IO sizes
        # pi_act=1
        # Data was written with apptag=0xFFFF
        # Reading the data back should disable all checking for Type 1 and 2
        "test_id": 317,
        "fio_opts": {
            "rw": 'read',
            "number_ios": NUMBER_IOS,
            "output-format": "json",
            "pi_act": 1,
            "apptag": "0x0000",
            "apptag_mask": "0xFFFF",
            },
        "pi_chk": "APPTAG,GUARD,REFTAG",
        "bs_low": BS_LOW,
        "bs_high": BS_HIGH,
        "skip": "type3",
        "test_class": DifDixTest,
    },
#
# Error cases related to block size and metadata size
#
    {
        # Use a min block size that is not a multiple of lba/elba size to
        # trigger an error.
        "test_id": 401,
        "fio_opts": {
            "rw": 'read',
            "number_ios": NUMBER_IOS,
            "pi_act": 0,
            "apptag": "0x8888",
            "apptag_mask": "0x0FFF",
            },
        "pi_chk": "APPTAG,GUARD,REFTAG",
        "bs_low": BS_LOW+0.5,
        "bs_high": BS_HIGH,
        "success": SUCCESS_NONZERO,
        "test_class": DifDixTest,
    },
    {
        # Use metadata size that is too small
        "test_id": 402,
        "fio_opts": {
            "rw": 'read',
            "number_ios": NUMBER_IOS,
            "pi_act": 0,
            "apptag": "0x8888",
            "apptag_mask": "0x0FFF",
            },
        "pi_chk": "APPTAG,GUARD,REFTAG",
        "bs_low": BS_LOW,
        "bs_high": BS_HIGH,
        "mdsize_adjustment": -1,
        "success": SUCCESS_NONZERO,
        "skip": "elba",
        "test_class": DifDixTest,
    },
    {
        # Read workload with variable IO sizes
        # pi_act=0
        # Should still work even if metadata size is too large
        "test_id": 403,
        "fio_opts": {
            "rw": 'read',
            "number_ios": NUMBER_IOS,
            "pi_act": 0,
            "apptag": "0x8888",
            "apptag_mask": "0x0FFF",
            },
        "pi_chk": "APPTAG,GUARD,REFTAG",
        "bs_low": BS_LOW,
        "bs_high": BS_HIGH,
        "mdsize_adjustment": 1,
        "test_class": DifDixTest,
    },
]


def parse_args():
    """Parse command-line arguments."""

    parser = argparse.ArgumentParser()
    parser.add_argument('-d', '--debug', help='Enable debug messages', action='store_true')
    parser.add_argument('-f', '--fio', help='path to file executable (e.g., ./fio)')
    parser.add_argument('-a', '--artifact-root', help='artifact root directory')
    parser.add_argument('-s', '--skip', nargs='+', type=int,
                        help='list of test(s) to skip')
    parser.add_argument('-o', '--run-only', nargs='+', type=int,
                        help='list of test(s) to run, skipping all others')
    parser.add_argument('--dut', help='target NVMe character device to test '
                        '(e.g., /dev/ng0n1). WARNING: THIS IS A DESTRUCTIVE TEST', required=True)
    parser.add_argument('-l', '--lbaf', nargs='+', type=int,
                        help='list of lba formats to test')
    parser.add_argument('-i', '--ioengine', default='io_uring_cmd')
    args = parser.parse_args()

    return args


def get_lbafs(args):
    """
    Determine which LBA formats to use. Use either the ones specified on the
    command line or if none are specified query the device and use all lba
    formats with metadata.
    """
    lbaf_list = []
    id_ns_cmd = f"sudo nvme id-ns --output-format=json {args.dut}".split(' ')
    id_ns_output = subprocess.check_output(id_ns_cmd)
    lbafs = json.loads(id_ns_output)['lbafs']
    if args.lbaf:
        for lbaf in args.lbaf:
            lbaf_list.append({'lbaf': lbaf, 'ds': 2 ** lbafs[lbaf]['ds'],
                              'ms': lbafs[lbaf]['ms'], })
            if lbafs[lbaf]['ms'] == 0:
                print(f'Error: lbaf {lbaf} has metadata size zero')
                sys.exit(1)
    else:
        for lbaf_num, lbaf in enumerate(lbafs):
            if lbaf['ms'] != 0:
                lbaf_list.append({'lbaf': lbaf_num, 'ds': 2 ** lbaf['ds'],
                                  'ms': lbaf['ms'], })

    return lbaf_list


def get_guard_pi(lbaf_list, args):
    """
    Find out how many bits of guard protection information are associated with
    each lbaf to be used. If this is not available assume 16-bit guard pi.
    Also record the bytes of protection information associated with the number
    of guard PI bits.
    """
    nvm_id_ns_cmd = f"sudo nvme nvm-id-ns --output-format=json {args.dut}".split(' ')
    try:
        nvm_id_ns_output = subprocess.check_output(nvm_id_ns_cmd)
    except subprocess.CalledProcessError:
        print(f"Non-zero return code from {' '.join(nvm_id_ns_cmd)}; " \
                "assuming all lbafs use 16b Guard Protection Information")
        for lbaf in lbaf_list:
            lbaf['guard_pi_bits'] = 16
    else:
        elbafs = json.loads(nvm_id_ns_output)['elbafs']
        for elbaf_num, elbaf in enumerate(elbafs):
            for lbaf in lbaf_list:
                if lbaf['lbaf'] == elbaf_num:
                    lbaf['guard_pi_bits'] = 16 << elbaf['pif']

    # For 16b Guard Protection Information, the PI requires 8 bytes
    # For 32b and 64b Guard PI, the PI requires 16 bytes
    for lbaf in lbaf_list:
        if lbaf['guard_pi_bits'] == 16:
            lbaf['pi_bytes'] = 8
        else:
            lbaf['pi_bytes'] = 16


def get_capabilities(args):
    """
    Determine what end-to-end data protection features the device supports.
    """
    caps = { 'pil': [], 'pitype': [], 'elba': [] }
    id_ns_cmd = f"sudo nvme id-ns --output-format=json {args.dut}".split(' ')
    id_ns_output = subprocess.check_output(id_ns_cmd)
    id_ns_json = json.loads(id_ns_output)

    mc = id_ns_json['mc']
    if mc & 1:
        caps['elba'].append(1)
    if mc & 2:
        caps['elba'].append(0)

    dpc = id_ns_json['dpc']
    if dpc & 1:
        caps['pitype'].append(1)
    if dpc & 2:
        caps['pitype'].append(2)
    if dpc & 4:
        caps['pitype'].append(3)
    if dpc & 8:
        caps['pil'].append(1)
    if dpc & 16:
        caps['pil'].append(0)

    for _, value in caps.items():
        if len(value) == 0:
            logging.error("One or more end-to-end data protection features unsupported: %s", caps)
            sys.exit(-1)

    return caps


def format_device(args, lbaf, pitype, pil, elba):
    """
    Format device using specified lba format with specified pitype, pil, and
    elba values.
    """

    format_cmd = f"sudo nvme format {args.dut} --lbaf={lbaf['lbaf']} " \
                 f"--pi={pitype} --pil={pil} --ms={elba} --force"
    logging.debug("Format command: %s", format_cmd)
    format_cmd = format_cmd.split(' ')
    format_cmd_result = subprocess.run(format_cmd, capture_output=True, check=False,
                                       encoding=locale.getpreferredencoding())

    # Sometimes nvme-cli may format the device successfully but fail to
    # rescan the namespaces after the format. Continue if this happens but
    # abort if some other error occurs.
    if format_cmd_result.returncode != 0:
        if 'failed to rescan namespaces' not in format_cmd_result.stderr \
                or 'Success formatting namespace' not in format_cmd_result.stdout:
            logging.error(format_cmd_result.stdout)
            logging.error(format_cmd_result.stderr)
            print("Unable to format device; skipping this configuration")
            return False

    logging.debug(format_cmd_result.stdout)
    return True


def difdix_test(test_env, args, lbaf, pitype, elba):
    """
    Adjust test arguments based on values of lbaf, pitype, and elba.  Then run
    the tests.
    """
    for test in TEST_LIST:
        test['force_skip'] = False

        blocksize = lbaf['ds']
        # Set fio blocksize parameter at runtime
        # If we formatted the device in extended LBA mode (e.g., 520-byte
        # sectors), we usually need to add the lba data size and metadata size
        # together for fio's bs parameter. However, if pi_act == 1 and the
        # device is formatted so that the metadata is the same size as the PI,
        # then the device will take care of everything and the application
        # should just use regular power of 2 lba data size even when the device
        # is in extended lba mode.
        if elba:
            if not test['fio_opts']['pi_act'] or lbaf['ms'] != lbaf['pi_bytes']:
                blocksize += lbaf['ms']
            test['fio_opts']['md_per_io_size'] = 0
        else:
        # If we are using a separate buffer for metadata, fio doesn't need to
        # do anything when pi_act==1 and protection information size is equal to
        # metadata size since the device is taking care of it all. If either of
        # the two conditions do not hold, then we do need to allocate a
        # separate metadata buffer.
            if test['fio_opts']['pi_act'] and lbaf['ms'] == lbaf['pi_bytes']:
                test['fio_opts']['md_per_io_size'] = 0
            else:
                test['fio_opts']['md_per_io_size'] = lbaf['ms'] * test['bs_high']

        test['fio_opts']['bsrange'] = f"{blocksize * test['bs_low']}-{blocksize * test['bs_high']}"
        if 'mdsize_adjustment' in test:
            test['fio_opts']['md_per_io_size'] += test['mdsize_adjustment']

        # Set fio pi_chk parameter at runtime. If the device is formatted
        # with Type 3 protection information, this means that the reference
        # tag is not checked and I/O commands may throw an error if they
        # are submitted with the REFTAG bit set in pi_chk. Make sure fio
        # does not set pi_chk's REFTAG bit if the device is formatted with
        # Type 3 PI.
        if 'pi_chk' in test:
            if pitype == 3 and 'REFTAG' in test['pi_chk']:
                test['fio_opts']['pi_chk'] = test['pi_chk'].replace('REFTAG','')
                logging.debug("Type 3 PI: dropping REFTAG bit")
            else:
                test['fio_opts']['pi_chk'] = test['pi_chk']

        if 'skip' in test:
            if pitype == 3 and 'type3' in test['skip']:
                test['force_skip'] = True
                logging.debug("Type 3 PI: skipping test case")
            if elba and 'elba' in test['skip']:
                test['force_skip'] = True
                logging.debug("extended lba format: skipping test case")

        logging.debug("Test %d: pi_act=%d, bsrange=%s, md_per_io_size=%d", test['test_id'],
                      test['fio_opts']['pi_act'], test['fio_opts']['bsrange'],
                      test['fio_opts']['md_per_io_size'])

    return run_fio_tests(TEST_LIST, test_env, args)


def main():
    """
    Run tests using fio's io_uring_cmd ioengine to exercise end-to-end data
    protection capabilities.
    """

    args = parse_args()

    if args.debug:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)

    artifact_root = args.artifact_root if args.artifact_root else \
        f"nvmept_pi-test-{time.strftime('%Y%m%d-%H%M%S')}"
    os.mkdir(artifact_root)
    print(f"Artifact directory is {artifact_root}")

    if args.fio:
        fio_path = str(Path(args.fio).absolute())
    else:
        fio_path = 'fio'
    print(f"fio path is {fio_path}")

    lbaf_list = get_lbafs(args)
    get_guard_pi(lbaf_list, args)
    caps = get_capabilities(args)
    print("Device capabilities:", caps)

    for test in TEST_LIST:
        test['fio_opts']['filename'] = args.dut
        test['fio_opts']['ioengine'] = args.ioengine

    test_env = {
              'fio_path': fio_path,
              'fio_root': str(Path(__file__).absolute().parent.parent),
              'artifact_root': artifact_root,
              'basename': 'nvmept_pi',
              }

    total = { 'passed':  0, 'failed': 0, 'skipped': 0 }

    try:
        for lbaf, pil, pitype, elba in itertools.product(lbaf_list, caps['pil'], caps['pitype'],
                                                         caps['elba']):
            print(f"\nlbaf: {lbaf}, pil: {pil}, pitype: {pitype}, elba: {elba}")

            if not format_device(args, lbaf, pitype, pil, elba):
                continue

            test_env['artifact_root'] = \
                os.path.join(artifact_root, f"lbaf{lbaf['lbaf']}pil{pil}pitype{pitype}" \
                    f"elba{elba}")
            os.mkdir(test_env['artifact_root'])

            passed, failed, skipped = difdix_test(test_env, args, lbaf, pitype, elba)

            total['passed'] += passed
            total['failed'] += failed
            total['skipped'] += skipped
    except KeyboardInterrupt:
        pass

    print(f"\n\n{total['passed']} test(s) passed, {total['failed']} failed, " \
            f"{total['skipped']} skipped")
    sys.exit(total['failed'])


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0-only
#
# Copyright (c) 2019 Western Digital Corporation or its affiliates.

"""
# readonly.py
#
# Do some basic tests of the --readonly parameter
#
# USAGE
# python readonly.py [-f fio-executable]
#
# EXAMPLES
# python t/readonly.py
# python t/readonly.py -f ./fio
#
# REQUIREMENTS
# Python 3.5+
#
"""

import os
import sys
import time
import argparse
from pathlib import Path
from fiotestlib import FioJobCmdTest, run_fio_tests
from fiotestcommon import SUCCESS_DEFAULT, SUCCESS_NONZERO


class FioReadOnlyTest(FioJobCmdTest):
    """fio read only test."""

    def setup(self, parameters):
        """Setup the test."""

        fio_args = [
                    "--name=readonly",
                    "--ioengine=null",
                    "--time_based",
                    "--runtime=1s",
                    "--size=1M",
                    f"--rw={self.fio_opts['rw']}",
                   ]
        if 'readonly-pre' in parameters:
            fio_args.insert(0, "--readonly")
        if 'readonly-post' in parameters:
            fio_args.append("--readonly")

        super().setup(fio_args)


TEST_LIST = [
            {
                "test_id": 1,
                "fio_opts": { "rw": "randread", },
                "readonly-pre": 1,
                "success": SUCCESS_DEFAULT,
                "test_class": FioReadOnlyTest,
            },
            {
                "test_id": 2,
                "fio_opts": { "rw": "randwrite", },
                "readonly-pre": 1,
                "success": SUCCESS_NONZERO,
                "test_class": FioReadOnlyTest,
            },
            {
                "test_id": 3,
                "fio_opts": { "rw": "randtrim", },
                "readonly-pre": 1,
                "success": SUCCESS_NONZERO,
                "test_class": FioReadOnlyTest,
            },
            {
                "test_id": 4,
                "fio_opts": { "rw": "randread", },
                "readonly-post": 1,
                "success": SUCCESS_DEFAULT,
                "test_class": FioReadOnlyTest,
            },
            {
                "test_id": 5,
                "fio_opts": { "rw": "randwrite", },
                "readonly-post": 1,
                "success": SUCCESS_NONZERO,
                "test_class": FioReadOnlyTest,
            },
            {
                "test_id": 6,
                "fio_opts": { "rw": "randtrim", },
                "readonly-post": 1,
                "success": SUCCESS_NONZERO,
                "test_class": FioReadOnlyTest,
            },
            {
                "test_id": 7,
                "fio_opts": { "rw": "randread", },
                "success": SUCCESS_DEFAULT,
                "test_class": FioReadOnlyTest,
            },
            {
                "test_id": 8,
                "fio_opts": { "rw": "randwrite", },
                "success": SUCCESS_DEFAULT,
                "test_class": FioReadOnlyTest,
            },
            {
                "test_id": 9,
                "fio_opts": { "rw": "randtrim", },
                "success": SUCCESS_DEFAULT,
                "test_class": FioReadOnlyTest,
            },
        ]


def parse_args():
    """Parse command-line arguments."""

    parser = argparse.ArgumentParser()
    parser.add_argument('-f', '--fio', help='path to fio executable (e.g., ./fio)')
    parser.add_argument('-a', '--artifact-root', help='artifact root directory')
    parser.add_argument('-s', '--skip', nargs='+', type=int,
                        help='list of test(s) to skip')
    parser.add_argument('-o', '--run-only', nargs='+', type=int,
                        help='list of test(s) to run, skipping all others')
    args = parser.parse_args()

    return args


def main():
    """Run readonly tests."""

    args = parse_args()

    if args.fio:
        fio_path = str(Path(args.fio).absolute())
    else:
        fio_path = 'fio'
    print(f"fio path is {fio_path}")

    artifact_root = args.artifact_root if args.artifact_root else \
        f"readonly-test-{time.strftime('%Y%m%d-%H%M%S')}"
    os.mkdir(artifact_root)
    print(f"Artifact directory is {artifact_root}")

    test_env = {
              'fio_path': fio_path,
              'fio_root': str(Path(__file__).absolute().parent.parent),
              'artifact_root': artifact_root,
              'basename': 'readonly',
              }

    _, failed, _ = run_fio_tests(TEST_LIST, test_env, args)
    sys.exit(failed)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0-only
#
# Copyright (c) 2020 Western Digital Corporation or its affiliates.
#
"""
# latency_percentiles.py
#
# Test the code that produces latency percentiles
# This is mostly to test the code changes to allow reporting
# of slat, clat, and lat percentiles
#
# USAGE
# python3 latency-tests.py [-f fio-path] [-a artifact-root] [--debug]
#
#
# Test scenarios:
#
# - DONE json
#   unified rw reporting
#   compare with latency log
#   try various combinations of the ?lat_percentile options
#   null, aio
#   r, w, t
# - DONE json+
#   check presence of latency bins
#   if the json percentiles match those from the raw data
#   then the latency bin values and counts are probably ok
# - DONE terse
#   produce both terse, JSON output and confirm that they match
#   lat only; both lat and clat
# - DONE sync_lat
#   confirm that sync_lat data appears
# - MANUAL TESTING normal output:
#       null ioengine
#           enable all, but only clat and lat appear
#           enable subset of latency types
#           read, write, trim, unified
#       libaio ioengine
#           enable all latency types
#           enable subset of latency types
#           read, write, trim, unified
# ./fio/fio --name=test --randrepeat=0 --norandommap --time_based --runtime=2s --size=512M \
#       --ioengine=null --slat_percentiles=1 --clat_percentiles=1 --lat_percentiles=1
# echo confirm that clat and lat percentiles appear
# ./fio/fio --name=test --randrepeat=0 --norandommap --time_based --runtime=2s --size=512M \
#       --ioengine=null --slat_percentiles=0 --clat_percentiles=0 --lat_percentiles=1
# echo confirm that only lat percentiles appear
# ./fio/fio --name=test --randrepeat=0 --norandommap --time_based --runtime=2s --size=512M \
#       --ioengine=null --slat_percentiles=0 --clat_percentiles=1 --lat_percentiles=0
# echo confirm that only clat percentiles appear
# ./fio/fio --name=test --randrepeat=0 --norandommap --time_based --runtime=2s --size=512M \
#       --ioengine=libaio --slat_percentiles=1 --clat_percentiles=1 --lat_percentiles=1
# echo confirm that slat, clat, lat percentiles appear
# ./fio/fio --name=test --randrepeat=0 --norandommap --time_based --runtime=2s --size=512M \
#       --ioengine=libaio --slat_percentiles=0 --clat_percentiles=1 --lat_percentiles=1
# echo confirm that clat and lat percentiles appear
# ./fio/fio --name=test --randrepeat=0 --norandommap --time_based --runtime=2s --size=512M \
#       --ioengine=libaio -rw=randrw
# echo confirm that clat percentiles appear for reads and writes
# ./fio/fio --name=test --randrepeat=0 --norandommap --time_based --runtime=2s --size=512M \
#       --ioengine=libaio --slat_percentiles=1 --clat_percentiles=0 --lat_percentiles=0 --rw=randrw
# echo confirm that slat percentiles appear for both reads and writes
# ./fio/fio --name=test --randrepeat=0 --norandommap --time_based --runtime=2s --size=512M \
#       --ioengine=libaio --slat_percentiles=1 --clat_percentiles=1 --lat_percentiles=1 \
#       --rw=randrw --unified_rw_reporting=1
# echo confirm that slat, clat, and lat percentiles appear for 'mixed' IOs
#./fio/fio --name=test --randrepeat=0 --norandommap --time_based --runtime=2s --size=512M \
#       --ioengine=null --slat_percentiles=1 --clat_percentiles=1 --lat_percentiles=1 \
#       --rw=randrw --fsync=32
# echo confirm that fsync latencies appear
"""

import os
import csv
import sys
import json
import math
import time
import argparse
import platform
import subprocess
from collections import Counter
from pathlib import Path


class FioLatTest():
    """fio latency percentile test."""

    def __init__(self, artifact_root, test_options, debug):
        """
        artifact_root   root directory for artifacts (subdirectory will be created under here)
        test            test specification
        """
        self.artifact_root = artifact_root
        self.test_options = test_options
        self.debug = debug
        self.filename = None
        self.json_data = None
        self.terse_data = None

        self.test_dir = os.path.join(self.artifact_root,
                                     "{:03d}".format(self.test_options['test_id']))
        if not os.path.exists(self.test_dir):
            os.mkdir(self.test_dir)

        self.filename = "latency{:03d}".format(self.test_options['test_id'])

    def run_fio(self, fio_path):
        """Run a test."""

        fio_args = [
            "--max-jobs=16",
            "--name=latency",
            "--randrepeat=0",
            "--norandommap",
            "--time_based",
            "--size=16M",
            "--rwmixread=50",
            "--group_reporting=1",
            "--write_lat_log={0}".format(self.filename),
            "--output={0}.out".format(self.filename),
            "--ioengine={ioengine}".format(**self.test_options),
            "--rw={rw}".format(**self.test_options),
            "--runtime={runtime}".format(**self.test_options),
            "--output-format={output-format}".format(**self.test_options),
        ]
        for opt in ['slat_percentiles', 'clat_percentiles', 'lat_percentiles',
                    'unified_rw_reporting', 'fsync', 'fdatasync', 'numjobs',
                    'cmdprio_percentage', 'bssplit', 'cmdprio_bssplit']:
            if opt in self.test_options:
                option = '--{0}={{{0}}}'.format(opt)
                fio_args.append(option.format(**self.test_options))

        command = [fio_path] + fio_args
        with open(os.path.join(self.test_dir, "{0}.command".format(self.filename)), "w+") as \
                command_file:
            command_file.write("%s\n" % command)

        passed = True
        stdout_file = open(os.path.join(self.test_dir, "{0}.stdout".format(self.filename)), "w+")
        stderr_file = open(os.path.join(self.test_dir, "{0}.stderr".format(self.filename)), "w+")
        exitcode_file = open(os.path.join(self.test_dir,
                                          "{0}.exitcode".format(self.filename)), "w+")
        try:
            proc = None
            # Avoid using subprocess.run() here because when a timeout occurs,
            # fio will be stopped with SIGKILL. This does not give fio a
            # chance to clean up and means that child processes may continue
            # running and submitting IO.
            proc = subprocess.Popen(command,
                                    stdout=stdout_file,
                                    stderr=stderr_file,
                                    cwd=self.test_dir,
                                    universal_newlines=True)
            proc.communicate(timeout=300)
            exitcode_file.write('{0}\n'.format(proc.returncode))
            passed &= (proc.returncode == 0)
        except subprocess.TimeoutExpired:
            proc.terminate()
            proc.communicate()
            assert proc.poll()
            print("Timeout expired")
            passed = False
        except Exception:
            if proc:
                if not proc.poll():
                    proc.terminate()
                    proc.communicate()
            print("Exception: %s" % sys.exc_info())
            passed = False
        finally:
            stdout_file.close()
            stderr_file.close()
            exitcode_file.close()

        if passed:
            if 'json' in self.test_options['output-format']:
                if not self.get_json():
                    print('Unable to decode JSON data')
                    passed = False
            if 'terse' in self.test_options['output-format']:
                if not self.get_terse():
                    print('Unable to decode terse data')
                    passed = False

        return passed

    def get_json(self):
        """Convert fio JSON output into a python JSON object"""

        filename = os.path.join(self.test_dir, "{0}.out".format(self.filename))
        with open(filename, 'r') as file:
            file_data = file.read()

        #
        # Sometimes fio informational messages are included at the top of the
        # JSON output, especially under Windows. Try to decode output as JSON
        # data, lopping off up to the first four lines
        #
        lines = file_data.splitlines()
        for i in range(5):
            file_data = '\n'.join(lines[i:])
            try:
                self.json_data = json.loads(file_data)
            except json.JSONDecodeError:
                continue
            else:
                return True

        return False

    def get_terse(self):
        """Read fio output and return terse format data."""

        filename = os.path.join(self.test_dir, "{0}.out".format(self.filename))
        with open(filename, 'r') as file:
            file_data = file.read()

        #
        # Read the first few lines and see if any of them begin with '3;'
        # If so, the line is probably terse output. Obviously, this only
        # works for fio terse version 3 and it does not work for
        # multi-line terse output
        #
        lines = file_data.splitlines()
        for i in range(8):
            file_data = lines[i]
            if file_data.startswith('3;'):
                self.terse_data = file_data.split(';')
                return True

        return False

    def check_latencies(self, jsondata, ddir, slat=True, clat=True, tlat=True, plus=False,
                        unified=False):
        """Check fio latency data.

        ddir                data direction to check (0=read, 1=write, 2=trim)
        slat                True if submission latency data available to check
        clat                True if completion latency data available to check
        tlat                True of total latency data available to check
        plus                True if we actually have json+ format data where additional checks can
                            be carried out
        unified             True if fio is reporting unified r/w data
        """

        types = {
            'slat': slat,
            'clat': clat,
            'lat': tlat
        }

        retval = True

        for lat in ['slat', 'clat', 'lat']:
            this_iter = True
            if not types[lat]:
                if 'percentile' in jsondata[lat+'_ns']:
                    this_iter = False
                    print('unexpected %s percentiles found' % lat)
                else:
                    print("%s percentiles skipped" % lat)
                continue
            else:
                if 'percentile' not in jsondata[lat+'_ns']:
                    this_iter = False
                    print('%s percentiles not found in fio output' % lat)

            #
            # Check only for the presence/absence of json+
            # latency bins. Future work can check the
            # accuracy of the bin values and counts.
            #
            # Because the latency percentiles are based on
            # the bins, we can be confident that the bin
            # values and counts are correct if fio's
            # latency percentiles match what we compute
            # from the raw data.
            #
            if plus:
                if 'bins' not in jsondata[lat+'_ns']:
                    print('bins not found with json+ output format')
                    this_iter = False
                else:
                    if not self.check_jsonplus(jsondata[lat+'_ns']):
                        this_iter = False
            else:
                if 'bins' in jsondata[lat+'_ns']:
                    print('json+ bins found with json output format')
                    this_iter = False

            latencies = []
            for i in range(10):
                lat_file = os.path.join(self.test_dir, "%s_%s.%s.log" % (self.filename, lat, i+1))
                if not os.path.exists(lat_file):
                    break
                with open(lat_file, 'r', newline='') as file:
                    reader = csv.reader(file)
                    for line in reader:
                        if unified or int(line[2]) == ddir:
                            latencies.append(int(line[1]))

            if int(jsondata['total_ios']) != len(latencies):
                this_iter = False
                print('%s: total_ios = %s, latencies logged = %d' % \
                        (lat, jsondata['total_ios'], len(latencies)))
            elif self.debug:
                print("total_ios %s match latencies logged" % jsondata['total_ios'])

            latencies.sort()
            ptiles = jsondata[lat+'_ns']['percentile']

            for percentile in ptiles.keys():
                #
                # numpy.percentile(latencies, float(percentile),
                #       interpolation='higher')
                # produces values that mostly match what fio reports
                # however, in the tails of the distribution, the values produced
                # by fio's and numpy.percentile's algorithms are occasionally off
                # by one latency measurement. So instead of relying on the canned
                # numpy.percentile routine, implement here fio's algorithm
                #
                rank = math.ceil(float(percentile)/100 * len(latencies))
                if rank > 0:
                    index = rank - 1
                else:
                    index = 0
                value = latencies[int(index)]
                fio_val = int(ptiles[percentile])
                # The theory in stat.h says that the proportional error will be
                # less than 1/128
                if not self.similar(fio_val, value):
                    delta = abs(fio_val - value) / value
                    print("Error with %s %sth percentile: "
                          "fio: %d, expected: %d, proportional delta: %f" %
                          (lat, percentile, fio_val, value, delta))
                    print("Rank: %d, index: %d" % (rank, index))
                    this_iter = False
                elif self.debug:
                    print('%s %sth percentile values match: %d, %d' %
                          (lat, percentile, fio_val, value))

            if this_iter:
                print("%s percentiles match" % lat)
            else:
                retval = False

        return retval

    @staticmethod
    def check_empty(job):
        """
        Make sure JSON data is empty.

        Some data structures should be empty. This function makes sure that they are.

        job         JSON object that we need to check for emptiness
        """

        return job['total_ios'] == 0 and \
                job['slat_ns']['N'] == 0 and \
                job['clat_ns']['N'] == 0 and \
                job['lat_ns']['N'] == 0

    def check_nocmdprio_lat(self, job):
        """
        Make sure no per priority latencies appear.

        job         JSON object to check
        """

        for ddir in ['read', 'write', 'trim']:
            if ddir in job:
                if 'prios' in job[ddir]:
                    print("Unexpected per priority latencies found in %s output" % ddir)
                    return False

        if self.debug:
            print("No per priority latencies found")

        return True

    @staticmethod
    def similar(approximation, actual):
        """
        Check whether the approximate values recorded by fio are within the theoretical bound.

        Since it is impractical to store exact latency measurements for each and every IO, fio
        groups similar latency measurements into variable-sized bins. The theory in stat.h says
        that the proportional error will be less than 1/128. This function checks whether this
        is true.

        TODO This test will fail when comparing a value from the largest latency bin against its
        actual measurement. Find some way to detect this and avoid failing.

        approximation   value of the bin used by fio to store a given latency
        actual          actual latency value
        """

        # Avoid a division by zero. The smallest latency values have no error.
        if actual == 0:
            return approximation == 0

        delta = abs(approximation - actual) / actual
        return delta <= 1/128

    def check_jsonplus(self, jsondata):
        """Check consistency of json+ data

        When we have json+ data we can check the min value, max value, and
        sample size reported by fio

        jsondata            json+ data that we need to check
        """

        retval = True

        keys = [int(k) for k in jsondata['bins'].keys()]
        values = [int(jsondata['bins'][k]) for k in jsondata['bins'].keys()]
        smallest = min(keys)
        biggest = max(keys)
        sampsize = sum(values)

        if not self.similar(jsondata['min'], smallest):
            retval = False
            print('reported min %d does not match json+ min %d' % (jsondata['min'], smallest))
        elif self.debug:
            print('json+ min values match: %d' % jsondata['min'])

        if not self.similar(jsondata['max'], biggest):
            retval = False
            print('reported max %d does not match json+ max %d' % (jsondata['max'], biggest))
        elif self.debug:
            print('json+ max values match: %d' % jsondata['max'])

        if sampsize != jsondata['N']:
            retval = False
            print('reported sample size %d does not match json+ total count %d' % \
                    (jsondata['N'], sampsize))
        elif self.debug:
            print('json+ sample sizes match: %d' % sampsize)

        return retval

    def check_sync_lat(self, jsondata, plus=False):
        """Check fsync latency percentile data.

        All we can check is that some percentiles are reported, unless we have json+ data.
        If we actually have json+ data then we can do more checking.

        jsondata        JSON data for fsync operations
        plus            True if we actually have json+ data
        """
        retval = True

        if 'percentile' not in jsondata['lat_ns']:
            print("Sync percentile data not found")
            return False

        if int(jsondata['total_ios']) != int(jsondata['lat_ns']['N']):
            retval = False
            print('Mismatch between total_ios and lat_ns sample size')
        elif self.debug:
            print('sync sample sizes match: %d' % jsondata['total_ios'])

        if not plus:
            if 'bins' in jsondata['lat_ns']:
                print('Unexpected json+ bin data found')
                return False

        if not self.check_jsonplus(jsondata['lat_ns']):
            retval = False

        return retval

    def check_terse(self, terse, jsondata):
        """Compare terse latencies with JSON latencies.

        terse           terse format data for checking
        jsondata        JSON format data for checking
        """

        retval = True

        for lat in terse:
            split = lat.split('%')
            pct = split[0]
            terse_val = int(split[1][1:])
            json_val = math.floor(jsondata[pct]/1000)
            if terse_val != json_val:
                retval = False
                print('Mismatch with %sth percentile: json value=%d,%d terse value=%d' % \
                        (pct, jsondata[pct], json_val, terse_val))
            elif self.debug:
                print('Terse %sth percentile matches JSON value: %d' % (pct, terse_val))

        return retval

    def check_prio_latencies(self, jsondata, clat=True, plus=False):
        """Check consistency of per priority latencies.

        clat                True if we should check clat data; other check lat data
        plus                True if we have json+ format data where additional checks can
                            be carried out
        unified             True if fio is reporting unified r/w data
        """

        if clat:
            obj = combined = 'clat_ns'
        else:
            obj = combined = 'lat_ns'

        if not 'prios' in jsondata or not combined in jsondata:
            print("Error identifying per priority latencies")
            return False

        sum_sample_size = sum([x[obj]['N'] for x in jsondata['prios']])
        if sum_sample_size != jsondata[combined]['N']:
            print("Per prio sample size sum %d != combined sample size %d" %
                  (sum_sample_size, jsondata[combined]['N']))
            return False
        elif self.debug:
            print("Per prio sample size sum %d == combined sample size %d" %
                  (sum_sample_size, jsondata[combined]['N']))

        min_val = min([x[obj]['min'] for x in jsondata['prios']])
        if min_val != jsondata[combined]['min']:
            print("Min per prio min latency %d does not match min %d from combined data" %
                  (min_val, jsondata[combined]['min']))
            return False
        elif self.debug:
            print("Min per prio min latency %d matches min %d from combined data" %
                  (min_val, jsondata[combined]['min']))

        max_val = max([x[obj]['max'] for x in jsondata['prios']])
        if max_val != jsondata[combined]['max']:
            print("Max per prio max latency %d does not match max %d from combined data" %
                  (max_val, jsondata[combined]['max']))
            return False
        elif self.debug:
            print("Max per prio max latency %d matches max %d from combined data" %
                  (max_val, jsondata[combined]['max']))

        weighted_vals = [x[obj]['mean'] * x[obj]['N'] for x in jsondata['prios']]
        weighted_avg = sum(weighted_vals) / jsondata[combined]['N']
        delta = abs(weighted_avg - jsondata[combined]['mean'])
        if (delta / jsondata[combined]['mean']) > 0.0001:
            print("Difference between merged per prio weighted average %f mean "
                  "and actual mean %f exceeds 0.01%%" % (weighted_avg, jsondata[combined]['mean']))
            return False
        elif self.debug:
            print("Merged per prio weighted average %f mean matches actual mean %f" %
                  (weighted_avg, jsondata[combined]['mean']))

        if plus:
            for prio in jsondata['prios']:
                if not self.check_jsonplus(prio[obj]):
                    return False

            counter = Counter()
            for prio in jsondata['prios']:
                counter.update(prio[obj]['bins'])

            bins = dict(counter)

            if len(bins) != len(jsondata[combined]['bins']):
                print("Number of merged bins %d does not match number of overall bins %d" %
                      (len(bins), len(jsondata[combined]['bins'])))
                return False
            elif self.debug:
                print("Number of merged bins %d matches number of overall bins %d" %
                      (len(bins), len(jsondata[combined]['bins'])))

            for duration in bins.keys():
                if bins[duration] != jsondata[combined]['bins'][duration]:
                    print("Merged per prio count does not match overall count for duration %d" %
                          duration)
                    return False

        print("Merged per priority latency data match combined latency data")
        return True

    def check(self):
        """Check test output."""

        raise NotImplementedError()


class Test001(FioLatTest):
    """Test object for Test 1."""

    def check(self):
        """Check Test 1 output."""

        job = self.json_data['jobs'][0]

        retval = True
        if not self.check_empty(job['write']):
            print("Unexpected write data found in output")
            retval = False
        if not self.check_empty(job['trim']):
            print("Unexpected trim data found in output")
            retval = False
        if not self.check_nocmdprio_lat(job):
            print("Unexpected per priority latencies found")
            retval = False

        retval &= self.check_latencies(job['read'], 0, slat=False)

        return retval


class Test002(FioLatTest):
    """Test object for Test 2."""

    def check(self):
        """Check Test 2 output."""

        job = self.json_data['jobs'][0]

        retval = True
        if not self.check_empty(job['read']):
            print("Unexpected read data found in output")
            retval = False
        if not self.check_empty(job['trim']):
            print("Unexpected trim data found in output")
            retval = False
        if not self.check_nocmdprio_lat(job):
            print("Unexpected per priority latencies found")
            retval = False

        retval &= self.check_latencies(job['write'], 1, slat=False, clat=False)

        return retval


class Test003(FioLatTest):
    """Test object for Test 3."""

    def check(self):
        """Check Test 3 output."""

        job = self.json_data['jobs'][0]

        retval = True
        if not self.check_empty(job['read']):
            print("Unexpected read data found in output")
            retval = False
        if not self.check_empty(job['write']):
            print("Unexpected write data found in output")
            retval = False
        if not self.check_nocmdprio_lat(job):
            print("Unexpected per priority latencies found")
            retval = False

        retval &= self.check_latencies(job['trim'], 2, slat=False, tlat=False)

        return retval


class Test004(FioLatTest):
    """Test object for Tests 4, 13."""

    def check(self):
        """Check Test 4, 13 output."""

        job = self.json_data['jobs'][0]

        retval = True
        if not self.check_empty(job['write']):
            print("Unexpected write data found in output")
            retval = False
        if not self.check_empty(job['trim']):
            print("Unexpected trim data found in output")
            retval = False
        if not self.check_nocmdprio_lat(job):
            print("Unexpected per priority latencies found")
            retval = False

        retval &= self.check_latencies(job['read'], 0, plus=True)

        return retval


class Test005(FioLatTest):
    """Test object for Test 5."""

    def check(self):
        """Check Test 5 output."""

        job = self.json_data['jobs'][0]

        retval = True
        if not self.check_empty(job['read']):
            print("Unexpected read data found in output")
            retval = False
        if not self.check_empty(job['trim']):
            print("Unexpected trim data found in output")
            retval = False
        if not self.check_nocmdprio_lat(job):
            print("Unexpected per priority latencies found")
            retval = False

        retval &= self.check_latencies(job['write'], 1, slat=False, plus=True)

        return retval


class Test006(FioLatTest):
    """Test object for Test 6."""

    def check(self):
        """Check Test 6 output."""

        job = self.json_data['jobs'][0]

        retval = True
        if not self.check_empty(job['write']):
            print("Unexpected write data found in output")
            retval = False
        if not self.check_empty(job['trim']):
            print("Unexpected trim data found in output")
            retval = False
        if not self.check_nocmdprio_lat(job):
            print("Unexpected per priority latencies found")
            retval = False

        retval &= self.check_latencies(job['read'], 0, slat=False, tlat=False, plus=True)

        return retval


class Test007(FioLatTest):
    """Test object for Test 7."""

    def check(self):
        """Check Test 7 output."""

        job = self.json_data['jobs'][0]

        retval = True
        if not self.check_empty(job['trim']):
            print("Unexpected trim data found in output")
            retval = False
        if not self.check_nocmdprio_lat(job):
            print("Unexpected per priority latencies found")
            retval = False

        retval &= self.check_latencies(job['read'], 0, clat=False, tlat=False, plus=True)
        retval &= self.check_latencies(job['write'], 1, clat=False, tlat=False, plus=True)

        return retval


class Test008(FioLatTest):
    """Test object for Tests 8, 14."""

    def check(self):
        """Check Test 8, 14 output."""

        job = self.json_data['jobs'][0]

        retval = True
        if 'read' in job or 'write' in job or 'trim' in job:
            print("Unexpected data direction found in fio output")
            retval = False
        if not self.check_nocmdprio_lat(job):
            print("Unexpected per priority latencies found")
            retval = False

        retval &= self.check_latencies(job['mixed'], 0, plus=True, unified=True)

        return retval


class Test009(FioLatTest):
    """Test object for Test 9."""

    def check(self):
        """Check Test 9 output."""

        job = self.json_data['jobs'][0]

        retval = True
        if not self.check_empty(job['read']):
            print("Unexpected read data found in output")
            retval = False
        if not self.check_empty(job['trim']):
            print("Unexpected trim data found in output")
            retval = False
        if not self.check_sync_lat(job['sync'], plus=True):
            print("Error checking fsync latency data")
            retval = False
        if not self.check_nocmdprio_lat(job):
            print("Unexpected per priority latencies found")
            retval = False

        retval &= self.check_latencies(job['write'], 1, slat=False, plus=True)

        return retval


class Test010(FioLatTest):
    """Test object for Test 10."""

    def check(self):
        """Check Test 10 output."""

        job = self.json_data['jobs'][0]

        retval = True
        if not self.check_empty(job['trim']):
            print("Unexpected trim data found in output")
            retval = False
        if not self.check_nocmdprio_lat(job):
            print("Unexpected per priority latencies found")
            retval = False

        retval &= self.check_latencies(job['read'], 0, plus=True)
        retval &= self.check_latencies(job['write'], 1, plus=True)
        retval &= self.check_terse(self.terse_data[17:34], job['read']['lat_ns']['percentile'])
        retval &= self.check_terse(self.terse_data[58:75], job['write']['lat_ns']['percentile'])
        # Terse data checking only works for default percentiles.
        # This needs to be changed if something other than the default is ever used.

        return retval


class Test011(FioLatTest):
    """Test object for Test 11."""

    def check(self):
        """Check Test 11 output."""

        job = self.json_data['jobs'][0]

        retval = True
        if not self.check_empty(job['trim']):
            print("Unexpected trim data found in output")
            retval = False
        if not self.check_nocmdprio_lat(job):
            print("Unexpected per priority latencies found")
            retval = False

        retval &= self.check_latencies(job['read'], 0, slat=False, clat=False, plus=True)
        retval &= self.check_latencies(job['write'], 1, slat=False, clat=False, plus=True)
        retval &= self.check_terse(self.terse_data[17:34], job['read']['lat_ns']['percentile'])
        retval &= self.check_terse(self.terse_data[58:75], job['write']['lat_ns']['percentile'])
        # Terse data checking only works for default percentiles.
        # This needs to be changed if something other than the default is ever used.

        return retval


class Test015(FioLatTest):
    """Test object for Test 15."""

    def check(self):
        """Check Test 15 output."""

        job = self.json_data['jobs'][0]

        retval = True
        if not self.check_empty(job['write']):
            print("Unexpected write data found in output")
            retval = False
        if not self.check_empty(job['trim']):
            print("Unexpected trim data found in output")
            retval = False

        retval &= self.check_latencies(job['read'], 0, plus=True)
        retval &= self.check_prio_latencies(job['read'], clat=False, plus=True)

        return retval


class Test016(FioLatTest):
    """Test object for Test 16."""

    def check(self):
        """Check Test 16 output."""

        job = self.json_data['jobs'][0]

        retval = True
        if not self.check_empty(job['read']):
            print("Unexpected read data found in output")
            retval = False
        if not self.check_empty(job['trim']):
            print("Unexpected trim data found in output")
            retval = False

        retval &= self.check_latencies(job['write'], 1, slat=False, plus=True)
        retval &= self.check_prio_latencies(job['write'], clat=False, plus=True)

        return retval


class Test017(FioLatTest):
    """Test object for Test 17."""

    def check(self):
        """Check Test 17 output."""

        job = self.json_data['jobs'][0]

        retval = True
        if not self.check_empty(job['write']):
            print("Unexpected write data found in output")
            retval = False
        if not self.check_empty(job['trim']):
            print("Unexpected trim data found in output")
            retval = False

        retval &= self.check_latencies(job['read'], 0, slat=False, tlat=False, plus=True)
        retval &= self.check_prio_latencies(job['read'], plus=True)

        return retval


class Test018(FioLatTest):
    """Test object for Test 18."""

    def check(self):
        """Check Test 18 output."""

        job = self.json_data['jobs'][0]

        retval = True
        if not self.check_empty(job['trim']):
            print("Unexpected trim data found in output")
            retval = False

        retval &= self.check_latencies(job['read'], 0, clat=False, tlat=False, plus=True)
        retval &= self.check_latencies(job['write'], 1, clat=False, tlat=False, plus=True)

        # We actually have json+ data but setting plus=False below avoids checking the
        # json+ bins which did not exist for clat and lat because this job is run with
        # clat_percentiles=0, lat_percentiles=0, However, we can still check the summary
        # statistics
        retval &= self.check_prio_latencies(job['write'], plus=False)
        retval &= self.check_prio_latencies(job['read'], plus=False)

        return retval


class Test019(FioLatTest):
    """Test object for Tests 19, 20."""

    def check(self):
        """Check Test 19, 20 output."""

        job = self.json_data['jobs'][0]

        retval = True
        if 'read' in job or 'write' in job or 'trim' in job:
            print("Unexpected data direction found in fio output")
            retval = False

        retval &= self.check_latencies(job['mixed'], 0, plus=True, unified=True)
        retval &= self.check_prio_latencies(job['mixed'], clat=False, plus=True)

        return retval


class Test021(FioLatTest):
    """Test object for Test 21."""

    def check(self):
        """Check Test 21 output."""

        job = self.json_data['jobs'][0]

        retval = True
        if not self.check_empty(job['trim']):
            print("Unexpected trim data found in output")
            retval = False

        retval &= self.check_latencies(job['read'], 0, slat=False, tlat=False, plus=True)
        retval &= self.check_latencies(job['write'], 1, slat=False, tlat=False, plus=True)
        retval &= self.check_prio_latencies(job['read'], clat=True, plus=True)
        retval &= self.check_prio_latencies(job['write'], clat=True, plus=True)

        return retval


def parse_args():
    """Parse command-line arguments."""

    parser = argparse.ArgumentParser()
    parser.add_argument('-f', '--fio', help='path to file executable (e.g., ./fio)')
    parser.add_argument('-a', '--artifact-root', help='artifact root directory')
    parser.add_argument('-d', '--debug', help='enable debug output', action='store_true')
    parser.add_argument('-s', '--skip', nargs='+', type=int,
                        help='list of test(s) to skip')
    parser.add_argument('-o', '--run-only', nargs='+', type=int,
                        help='list of test(s) to run, skipping all others')
    args = parser.parse_args()

    return args


def main():
    """Run tests of fio latency percentile reporting"""

    args = parse_args()

    artifact_root = args.artifact_root if args.artifact_root else \
        "latency-test-{0}".format(time.strftime("%Y%m%d-%H%M%S"))
    os.mkdir(artifact_root)
    print("Artifact directory is %s" % artifact_root)

    if args.fio:
        fio = str(Path(args.fio).absolute())
    else:
        fio = 'fio'
    print("fio path is %s" % fio)

    if platform.system() == 'Linux':
        aio = 'libaio'
    elif platform.system() == 'Windows':
        aio = 'windowsaio'
    else:
        aio = 'posixaio'

    test_list = [
        {
            # randread, null
            # enable slat, clat, lat
            # only clat and lat will appear because
            # because the null ioengine is synchronous
            "test_id": 1,
            "runtime": 2,
            "output-format": "json",
            "slat_percentiles": 1,
            "clat_percentiles": 1,
            "lat_percentiles": 1,
            "ioengine": 'null',
            'rw': 'randread',
            "test_obj": Test001,
        },
        {
            # randwrite, null
            # enable lat only
            "test_id": 2,
            "runtime": 2,
            "output-format": "json",
            "slat_percentiles": 0,
            "clat_percentiles": 0,
            "lat_percentiles": 1,
            "ioengine": 'null',
            'rw': 'randwrite',
            "test_obj": Test002,
        },
        {
            # randtrim, null
            # enable clat only
            "test_id": 3,
            "runtime": 2,
            "output-format": "json",
            "slat_percentiles": 0,
            "clat_percentiles": 1,
            "lat_percentiles": 0,
            "ioengine": 'null',
            'rw': 'randtrim',
            "test_obj": Test003,
        },
        {
            # randread, aio
            # enable slat, clat, lat
            # all will appear because libaio is asynchronous
            "test_id": 4,
            "runtime": 5,
            "output-format": "json+",
            "slat_percentiles": 1,
            "clat_percentiles": 1,
            "lat_percentiles": 1,
            "ioengine": aio,
            'rw': 'randread',
            "test_obj": Test004,
        },
        {
            # randwrite, aio
            # enable only clat, lat
            "test_id": 5,
            "runtime": 5,
            "output-format": "json+",
            "slat_percentiles": 0,
            "clat_percentiles": 1,
            "lat_percentiles": 1,
            "ioengine": aio,
            'rw': 'randwrite',
            "test_obj": Test005,
        },
        {
            # randread, aio
            # by default only clat should appear
            "test_id": 6,
            "runtime": 5,
            "output-format": "json+",
            "ioengine": aio,
            'rw': 'randread',
            "test_obj": Test006,
        },
        {
            # 50/50 r/w, aio
            # enable only slat
            "test_id": 7,
            "runtime": 5,
            "output-format": "json+",
            "slat_percentiles": 1,
            "clat_percentiles": 0,
            "lat_percentiles": 0,
            "ioengine": aio,
            'rw': 'randrw',
            "test_obj": Test007,
        },
        {
            # 50/50 r/w, aio, unified_rw_reporting
            # enable slat, clat, lat
            "test_id": 8,
            "runtime": 5,
            "output-format": "json+",
            "slat_percentiles": 1,
            "clat_percentiles": 1,
            "lat_percentiles": 1,
            "ioengine": aio,
            'rw': 'randrw',
            'unified_rw_reporting': 1,
            "test_obj": Test008,
        },
        {
            # randwrite, null
            # enable slat, clat, lat
            # fsync
            "test_id": 9,
            "runtime": 2,
            "output-format": "json+",
            "slat_percentiles": 1,
            "clat_percentiles": 1,
            "lat_percentiles": 1,
            "ioengine": 'null',
            'rw': 'randwrite',
            'fsync': 32,
            "test_obj": Test009,
        },
        {
            # 50/50 r/w, aio
            # enable slat, clat, lat
            "test_id": 10,
            "runtime": 5,
            "output-format": "terse,json+",
            "slat_percentiles": 1,
            "clat_percentiles": 1,
            "lat_percentiles": 1,
            "ioengine": aio,
            'rw': 'randrw',
            "test_obj": Test010,
        },
        {
            # 50/50 r/w, aio
            # enable only lat
            "test_id": 11,
            "runtime": 5,
            "output-format": "terse,json+",
            "slat_percentiles": 0,
            "clat_percentiles": 0,
            "lat_percentiles": 1,
            "ioengine": aio,
            'rw': 'randrw',
            "test_obj": Test011,
        },
        {
            # randread, null
            # enable slat, clat, lat
            # only clat and lat will appear because
            # because the null ioengine is synchronous
            # same as Test 1 except add numjobs = 4 to test
            # sum_thread_stats() changes
            "test_id": 12,
            "runtime": 2,
            "output-format": "json",
            "slat_percentiles": 1,
            "clat_percentiles": 1,
            "lat_percentiles": 1,
            "ioengine": 'null',
            'rw': 'randread',
            'numjobs': 4,
            "test_obj": Test001,
        },
        {
            # randread, aio
            # enable slat, clat, lat
            # all will appear because libaio is asynchronous
            # same as Test 4 except add numjobs = 4 to test
            # sum_thread_stats() changes
            "test_id": 13,
            "runtime": 5,
            "output-format": "json+",
            "slat_percentiles": 1,
            "clat_percentiles": 1,
            "lat_percentiles": 1,
            "ioengine": aio,
            'rw': 'randread',
            'numjobs': 4,
            "test_obj": Test004,
        },
        {
            # 50/50 r/w, aio, unified_rw_reporting
            # enable slat, clat, lata
            # same as Test 8 except add numjobs = 4 to test
            # sum_thread_stats() changes
            "test_id": 14,
            "runtime": 5,
            "output-format": "json+",
            "slat_percentiles": 1,
            "clat_percentiles": 1,
            "lat_percentiles": 1,
            "ioengine": aio,
            'rw': 'randrw',
            'unified_rw_reporting': 1,
            'numjobs': 4,
            "test_obj": Test008,
        },
        {
            # randread, aio
            # enable slat, clat, lat
            # all will appear because libaio is asynchronous
            # same as Test 4 except add cmdprio_percentage
            "test_id": 15,
            "runtime": 5,
            "output-format": "json+",
            "slat_percentiles": 1,
            "clat_percentiles": 1,
            "lat_percentiles": 1,
            "ioengine": aio,
            'rw': 'randread',
            'cmdprio_percentage': 50,
            "test_obj": Test015,
        },
        {
            # randwrite, aio
            # enable only clat, lat
            # same as Test 5 except add cmdprio_percentage
            "test_id": 16,
            "runtime": 5,
            "output-format": "json+",
            "slat_percentiles": 0,
            "clat_percentiles": 1,
            "lat_percentiles": 1,
            "ioengine": aio,
            'rw': 'randwrite',
            'cmdprio_percentage': 50,
            "test_obj": Test016,
        },
        {
            # randread, aio
            # by default only clat should appear
            # same as Test 6 except add cmdprio_percentage
            "test_id": 17,
            "runtime": 5,
            "output-format": "json+",
            "ioengine": aio,
            'rw': 'randread',
            'cmdprio_percentage': 50,
            "test_obj": Test017,
        },
        {
            # 50/50 r/w, aio
            # enable only slat
            # same as Test 7 except add cmdprio_percentage
            "test_id": 18,
            "runtime": 5,
            "output-format": "json+",
            "slat_percentiles": 1,
            "clat_percentiles": 0,
            "lat_percentiles": 0,
            "ioengine": aio,
            'rw': 'randrw',
            'cmdprio_percentage': 50,
            "test_obj": Test018,
        },
        {
            # 50/50 r/w, aio, unified_rw_reporting
            # enable slat, clat, lat
            # same as Test 8 except add cmdprio_percentage
            "test_id": 19,
            "runtime": 5,
            "output-format": "json+",
            "slat_percentiles": 1,
            "clat_percentiles": 1,
            "lat_percentiles": 1,
            "ioengine": aio,
            'rw': 'randrw',
            'unified_rw_reporting': 1,
            'cmdprio_percentage': 50,
            "test_obj": Test019,
        },
        {
            # 50/50 r/w, aio, unified_rw_reporting
            # enable slat, clat, lat
            # same as Test 19 except add numjobs = 4 to test
            # sum_thread_stats() changes
            "test_id": 20,
            "runtime": 5,
            "output-format": "json+",
            "slat_percentiles": 1,
            "clat_percentiles": 1,
            "lat_percentiles": 1,
            "ioengine": aio,
            'rw': 'randrw',
            'unified_rw_reporting': 1,
            'cmdprio_percentage': 50,
            'numjobs': 4,
            "test_obj": Test019,
        },
        {
            # r/w, aio
            # enable only clat
            # test bssplit and cmdprio_bssplit
            "test_id": 21,
            "runtime": 5,
            "output-format": "json+",
            "slat_percentiles": 0,
            "clat_percentiles": 1,
            "lat_percentiles": 0,
            "ioengine": aio,
            'rw': 'randrw',
            'bssplit': '64k/40:1024k/60',
            'cmdprio_bssplit': '64k/25/1/1:64k/75/3/2:1024k/0',
            "test_obj": Test021,
        },
        {
            # r/w, aio
            # enable only clat
            # same as Test 21 except add numjobs = 4 to test
            # sum_thread_stats() changes
            "test_id": 22,
            "runtime": 5,
            "output-format": "json+",
            "slat_percentiles": 0,
            "clat_percentiles": 1,
            "lat_percentiles": 0,
            "ioengine": aio,
            'rw': 'randrw',
            'bssplit': '64k/40:1024k/60',
            'cmdprio_bssplit': '64k/25/1/1:64k/75/3/2:1024k/0',
            'numjobs': 4,
            "test_obj": Test021,
        },
    ]

    passed = 0
    failed = 0
    skipped = 0

    for test in test_list:
        if (args.skip and test['test_id'] in args.skip) or \
           (args.run_only and test['test_id'] not in args.run_only):
            skipped = skipped + 1
            outcome = 'SKIPPED (User request)'
        elif (platform.system() != 'Linux' or os.geteuid() != 0) and \
             ('cmdprio_percentage' in test or 'cmdprio_bssplit' in test):
            skipped = skipped + 1
            outcome = 'SKIPPED (Linux root required for cmdprio tests)'
        else:
            test_obj = test['test_obj'](artifact_root, test, args.debug)
            status = test_obj.run_fio(fio)
            if status:
                status = test_obj.check()
            if status:
                passed = passed + 1
                outcome = 'PASSED'
            else:
                failed = failed + 1
                outcome = 'FAILED'

        print("**********Test {0} {1}**********".format(test['test_id'], outcome))

    print("{0} tests passed, {1} failed, {2} skipped".format(passed, failed, skipped))

    sys.exit(failed)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
"""
# nvmept.py
#
# Test fio's io_uring_cmd ioengine with NVMe pass-through commands.
#
# USAGE
# see python3 nvmept.py --help
#
# EXAMPLES
# python3 t/nvmept.py --dut /dev/ng0n1
# python3 t/nvmept.py --dut /dev/ng1n1 -f ./fio
#
# REQUIREMENTS
# Python 3.6
#
"""
import os
import sys
import time
import argparse
from pathlib import Path
from fiotestlib import FioJobCmdTest, run_fio_tests


class PassThruTest(FioJobCmdTest):
    """
    NVMe pass-through test class. Check to make sure output for selected data
    direction(s) is non-zero and that zero data appears for other directions.
    """

    def setup(self, parameters):
        """Setup a test."""

        fio_args = [
            "--name=nvmept",
            "--ioengine=io_uring_cmd",
            "--cmd_type=nvme",
            "--iodepth=8",
            "--iodepth_batch=4",
            "--iodepth_batch_complete=4",
            f"--filename={self.fio_opts['filename']}",
            f"--rw={self.fio_opts['rw']}",
            f"--output={self.filenames['output']}",
            f"--output-format={self.fio_opts['output-format']}",
        ]
        for opt in ['fixedbufs', 'nonvectored', 'force_async', 'registerfiles',
                    'sqthread_poll', 'sqthread_poll_cpu', 'hipri', 'nowait',
                    'time_based', 'runtime', 'verify', 'io_size']:
            if opt in self.fio_opts:
                option = f"--{opt}={self.fio_opts[opt]}"
                fio_args.append(option)

        super().setup(fio_args)


    def check_result(self):
        super().check_result()

        if 'rw' not in self.fio_opts:
            return

        if not self.passed:
            return

        job = self.json_data['jobs'][0]

        if self.fio_opts['rw'] in ['read', 'randread']:
            self.passed = self.check_all_ddirs(['read'], job)
        elif self.fio_opts['rw'] in ['write', 'randwrite']:
            if 'verify' not in self.fio_opts:
                self.passed = self.check_all_ddirs(['write'], job)
            else:
                self.passed = self.check_all_ddirs(['read', 'write'], job)
        elif self.fio_opts['rw'] in ['trim', 'randtrim']:
            self.passed = self.check_all_ddirs(['trim'], job)
        elif self.fio_opts['rw'] in ['readwrite', 'randrw']:
            self.passed = self.check_all_ddirs(['read', 'write'], job)
        elif self.fio_opts['rw'] in ['trimwrite', 'randtrimwrite']:
            self.passed = self.check_all_ddirs(['trim', 'write'], job)
        else:
            print(f"Unhandled rw value {self.fio_opts['rw']}")
            self.passed = False

        if job['iodepth_level']['8'] < 95:
            print("Did not achieve requested iodepth")
            self.passed = False


class FlushTest(FioJobCmdTest):
    def setup(self, parameters):
        fio_args = [
            "--name=nvmept-flush",
            "--ioengine=io_uring_cmd",
            "--cmd_type=nvme",
            "--randrepeat=0",
            f"--filename={self.fio_opts['filename']}",
            f"--rw={self.fio_opts['rw']}",
            f"--output={self.filenames['output']}",
            f"--output-format={self.fio_opts['output-format']}",
        ]

        for opt in ['fixedbufs', 'nonvectored', 'force_async', 'registerfiles',
                    'sqthread_poll', 'sqthread_poll_cpu', 'hipri', 'nowait',
                    'time_based', 'runtime', 'verify', 'io_size', 'num_range',
                    'iodepth', 'iodepth_batch', 'iodepth_batch_complete',
                    'size', 'rate', 'bs', 'bssplit', 'bsrange', 'randrepeat',
                    'buffer_pattern', 'verify_pattern', 'offset', 'fdp',
                    'fdp_pli', 'fdp_pli_select', 'dataplacement', 'plid_select',
                    'plids', 'dp_scheme', 'number_ios', 'read_iolog', 'fsync']:
            if opt in self.fio_opts:
                option = f"--{opt}={self.fio_opts[opt]}"
                fio_args.append(option)

        super().setup(fio_args)

    def check_result(self):
        super().check_result()

        job = self.json_data['jobs'][0]

        rw = self.fio_opts['rw']
        fsync = self.fio_opts['fsync']

        nr_write = job['write']['total_ios']
        nr_sync = job['sync']['total_ios']

        nr_sync_exp = nr_write // fsync

        # The actual number of DDIR_SYNC issued might miss one DDIR_SYNC command
        # when the last command issued was DDIR_WRITE command.
        if not ((nr_sync == nr_sync_exp) or (nr_sync + 1 == nr_sync_exp)):
            logging.error(f"nr_write={nr_write}, nr_sync={nr_sync}, fsync={fsync}")
            self.passed = False


TEST_LIST = [
    {
        "test_id": 1,
        "fio_opts": {
            "rw": 'read',
            "timebased": 1,
            "runtime": 3,
            "output-format": "json",
            },
        "test_class": PassThruTest,
    },
    {
        "test_id": 2,
        "fio_opts": {
            "rw": 'randread',
            "timebased": 1,
            "runtime": 3,
            "output-format": "json",
            },
        "test_class": PassThruTest,
    },
    {
        "test_id": 3,
        "fio_opts": {
            "rw": 'write',
            "timebased": 1,
            "runtime": 3,
            "output-format": "json",
            },
        "test_class": PassThruTest,
    },
    {
        "test_id": 4,
        "fio_opts": {
            "rw": 'randwrite',
            "timebased": 1,
            "runtime": 3,
            "output-format": "json",
            },
        "test_class": PassThruTest,
    },
    {
        "test_id": 5,
        "fio_opts": {
            "rw": 'trim',
            "timebased": 1,
            "runtime": 3,
            "output-format": "json",
            },
        "test_class": PassThruTest,
    },
    {
        "test_id": 6,
        "fio_opts": {
            "rw": 'randtrim',
            "timebased": 1,
            "runtime": 3,
            "output-format": "json",
            },
        "test_class": PassThruTest,
    },
    {
        "test_id": 7,
        "fio_opts": {
            "rw": 'write',
            "io_size": 1024*1024,
            "verify": "crc32c",
            "output-format": "json",
            },
        "test_class": PassThruTest,
    },
    {
        "test_id": 8,
        "fio_opts": {
            "rw": 'randwrite',
            "io_size": 1024*1024,
            "verify": "crc32c",
            "output-format": "json",
            },
        "test_class": PassThruTest,
    },
    {
        "test_id": 9,
        "fio_opts": {
            "rw": 'readwrite',
            "timebased": 1,
            "runtime": 3,
            "output-format": "json",
            },
        "test_class": PassThruTest,
    },
    {
        "test_id": 10,
        "fio_opts": {
            "rw": 'randrw',
            "timebased": 1,
            "runtime": 3,
            "output-format": "json",
            },
        "test_class": PassThruTest,
    },
    {
        "test_id": 11,
        "fio_opts": {
            "rw": 'trimwrite',
            "timebased": 1,
            "runtime": 3,
            "output-format": "json",
            },
        "test_class": PassThruTest,
    },
    {
        "test_id": 12,
        "fio_opts": {
            "rw": 'randtrimwrite',
            "timebased": 1,
            "runtime": 3,
            "output-format": "json",
            },
        "test_class": PassThruTest,
    },
    {
        "test_id": 13,
        "fio_opts": {
            "rw": 'randread',
            "timebased": 1,
            "runtime": 3,
            "fixedbufs": 1,
            "nonvectored": 1,
            "force_async": 1,
            "registerfiles": 1,
            "sqthread_poll": 1,
            "output-format": "json",
            },
        "test_class": PassThruTest,
    },
    {
        "test_id": 14,
        "fio_opts": {
            "rw": 'randwrite',
            "timebased": 1,
            "runtime": 3,
            "fixedbufs": 1,
            "nonvectored": 1,
            "force_async": 1,
            "registerfiles": 1,
            "sqthread_poll": 1,
            "output-format": "json",
            },
        "test_class": PassThruTest,
    },
    {
        # We can't enable fixedbufs because for trim-only
        # workloads fio actually does not allocate any buffers
        "test_id": 15,
        "fio_opts": {
            "rw": 'randtrim',
            "timebased": 1,
            "runtime": 3,
            "fixedbufs": 0,
            "nonvectored": 1,
            "force_async": 1,
            "registerfiles": 1,
            "sqthread_poll": 1,
            "output-format": "json",
            },
        "test_class": PassThruTest,
    },
    {
        "test_id": 16,
        "fio_opts": {
            "rw": 'read',
            "bs": 4096,
            "number_ios": 10,
            "fsync": 1,
            "output-format": "json",
            },
        "test_class": FlushTest,
    },
    {
        "test_id": 17,
        "fio_opts": {
            "rw": 'write',
            "bs": 4096,
            "number_ios": 10,
            "fsync": 1,
            "output-format": "json",
            },
        "test_class": FlushTest,
    },
    {
        "test_id": 18,
        "fio_opts": {
            "rw": 'readwrite',
            "bs": 4096,
            "number_ios": 10,
            "fsync": 1,
            "output-format": "json",
            },
        "test_class": FlushTest,
    },
    {
        "test_id": 19,
        "fio_opts": {
            "rw": 'trimwrite',
            "bs": 4096,
            "number_ios": 10,
            "fsync": 1,
            "output-format": "json",
            },
        "test_class": FlushTest,
    },
]

def parse_args():
    """Parse command-line arguments."""

    parser = argparse.ArgumentParser()
    parser.add_argument('-f', '--fio', help='path to file executable (e.g., ./fio)')
    parser.add_argument('-a', '--artifact-root', help='artifact root directory')
    parser.add_argument('-s', '--skip', nargs='+', type=int,
                        help='list of test(s) to skip')
    parser.add_argument('-o', '--run-only', nargs='+', type=int,
                        help='list of test(s) to run, skipping all others')
    parser.add_argument('--dut', help='target NVMe character device to test '
                        '(e.g., /dev/ng0n1). WARNING: THIS IS A DESTRUCTIVE TEST', required=True)
    args = parser.parse_args()

    return args


def main():
    """Run tests using fio's io_uring_cmd ioengine to send NVMe pass through commands."""

    args = parse_args()

    artifact_root = args.artifact_root if args.artifact_root else \
        f"nvmept-test-{time.strftime('%Y%m%d-%H%M%S')}"
    os.mkdir(artifact_root)
    print(f"Artifact directory is {artifact_root}")

    if args.fio:
        fio_path = str(Path(args.fio).absolute())
    else:
        fio_path = 'fio'
    print(f"fio path is {fio_path}")

    for test in TEST_LIST:
        test['fio_opts']['filename'] = args.dut

    test_env = {
              'fio_path': fio_path,
              'fio_root': str(Path(__file__).absolute().parent.parent),
              'artifact_root': artifact_root,
              'basename': 'nvmept',
              }

    _, failed, _ = run_fio_tests(TEST_LIST, test_env, args)
    sys.exit(failed)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# Note: this script is python2 and python 3 compatible.
#
# sgunmap-test.py
#
# Limited functionality test for trim workloads using fio's sg ioengine
# This checks only the three sets of reported iodepths
#
# !!!WARNING!!!
# This script carries out destructive tests. Be sure that
# there is no data you want to keep on the supplied devices.
#
# USAGE
# sgunmap-test.py char-device block-device fio-executable
#
# EXAMPLE
# t/sgunmap-test.py /dev/sg1 /dev/sdb ./fio
#
# REQUIREMENTS
# Python 2.6+
#
# TEST MATRIX
# For both char-dev and block-dev these are the expected
# submit/complete IO depths
#
#                       blockdev                chardev
#                       iodepth                 iodepth
# R QD1                 sub/comp: 1-4=100%      sub/comp: 1-4=100%
# W QD1                 sub/comp: 1-4=100%      sub/comp: 1-4=100%
# T QD1                 sub/comp: 1-4=100%      sub/comp: 1-4=100%
#
# R QD16, batch8        sub/comp: 1-4=100%      sub/comp: 1-4=100%
# W QD16, batch8        sub/comp: 1-4=100%      sub/comp: 1-4=100%
# T QD16, batch8        sub/comp: 1-4=100%      sub/comp: 5-8=100%
#
# R QD16, batch16       sub/comp: 1-4=100%      sub/comp: 1-4=100%
# W QD16, batch16       sub/comp: 1-4=100%      sub/comp: 1-4=100%
# T QD16, batch16       sub/comp: 1-4=100%      sub/comp: 9-16=100%
#

from __future__ import absolute_import
from __future__ import print_function
import sys
import json
import argparse
import traceback
import subprocess


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('chardev',
                        help='character device target (e.g., /dev/sg0)')
    parser.add_argument('blockdev',
                        help='block device target (e.g., /dev/sda)')
    parser.add_argument('fio',
                        help='path to fio executable (e.g., ./fio)')
    args = parser.parse_args()

    return args

#
# With block devices,
#     iodepth = 1 always
#     submit = complete = 1-4 always
# With character devices,
# RW
#     iodepth = qd
#     submit = 1-4
#     complete = 1-4 except for the IOs in flight
#                when the job is ending
# T
#     iodepth = qd
#     submit = qdbatch
#     complete = qdbatch except for the IOs in flight
#                when the job is ending
#


def check(jsondata, parameters, block, qd, qdbatch, rw):
    iodepth = jsondata['iodepth_level']
    submit = jsondata['iodepth_submit']
    complete = jsondata['iodepth_complete']

    try:
        if block:
            assert iodepth['1'] == 100.0
            assert submit['4'] == 100.0
            assert complete['4'] == 100.0
        elif 'read' in rw or 'write' in rw:
            assert iodepth[str(qd)] > 99.9
            assert submit['4'] == 100.0
            assert complete['4'] > 99.9
        else:
            if qdbatch <= 4:
                batchkey = '4'
            elif qdbatch > 64:
                batchkey = '>=64'
            else:
                batchkey = str(qdbatch)
            if qd >= 64:
                qdkey = ">=64"
            else:
                qdkey = str(qd)
            assert iodepth[qdkey] > 99
            assert submit[batchkey] == 100.0
            assert complete[batchkey] > 99
    except AssertionError:
        print("Assertion failed")
        traceback.print_exc()
        print(jsondata)
        return

    print("**********passed*********")


def runalltests(args, qd, batch):
    block = False
    for dev in [args.chardev, args.blockdev]:
        for rw in ["randread", "randwrite", "randtrim"]:
            parameters = ["--name=test",
                           "--time_based",
                           "--runtime=30s",
                           "--output-format=json",
                           "--ioengine=sg",
                           "--rw={0}".format(rw),
                           "--filename={0}".format(dev),
                           "--iodepth={0}".format(qd),
                           "--iodepth_batch={0}".format(batch)]

            print(parameters)
            output = subprocess.check_output([args.fio] + parameters)
            jsondata = json.loads(output)
            jobdata = jsondata['jobs'][0]
            check(jobdata, parameters, block, qd, batch, rw)
        block = True


def runcdevtrimtest(args, qd, batch):
    parameters = ["--name=test",
                   "--time_based",
                   "--runtime=30s",
                   "--output-format=json",
                   "--ioengine=sg",
                   "--rw=randtrim",
                   "--filename={0}".format(args.chardev),
                   "--iodepth={0}".format(qd),
                   "--iodepth_batch={0}".format(batch)]

    print(parameters)
    output = subprocess.check_output([args.fio] + parameters)
    jsondata = json.loads(output)
    jobdata = jsondata['jobs'][0]
    check(jobdata, parameters, False, qd, batch, "randtrim")


if __name__ == '__main__':
    args = parse_args()

    runcdevtrimtest(args, 32, 2)
    runcdevtrimtest(args, 32, 4)
    runcdevtrimtest(args, 32, 8)
    runcdevtrimtest(args, 64, 4)
    runcdevtrimtest(args, 64, 8)
    runcdevtrimtest(args, 64, 16)
    runcdevtrimtest(args, 128, 8)
    runcdevtrimtest(args, 128, 16)
    runcdevtrimtest(args, 128, 32)

    runalltests(args, 1, 1)
    runalltests(args, 16, 2)
    runalltests(args, 16, 16)



#!/usr/bin/env python3
#
# Copyright 2024 Samsung Electronics Co., Ltd All Rights Reserved
#
# For conditions of distribution and use, see the accompanying COPYING file.
#
"""
# nvmept_streams.py
#
# Test fio's NVMe streams support using the io_uring_cmd ioengine with NVMe
# pass-through commands.
#
# USAGE
# see python3 nvmept_streams.py --help
#
# EXAMPLES
# python3 t/nvmept_streams.py --dut /dev/ng0n1
# python3 t/nvmept_streams.py --dut /dev/ng1n1 -f ./fio
#
# REQUIREMENTS
# Python 3.6
#
# WARNING
# This is a destructive test
#
# Enable streams with
# nvme dir-send -D 0 -O 1 -e 1 -T 1 /dev/nvme0n1
#
# See streams directive status with
# nvme dir-receive -D 0 -O 1 -H /dev/nvme0n1
"""
import os
import sys
import time
import locale
import logging
import argparse
import subprocess
from pathlib import Path
from fiotestlib import FioJobCmdTest, run_fio_tests
from fiotestcommon import SUCCESS_NONZERO


class StreamsTest(FioJobCmdTest):
    """
    NVMe pass-through test class for streams. Check to make sure output for
    selected data direction(s) is non-zero and that zero data appears for other
    directions.
    """

    def setup(self, parameters):
        """Setup a test."""

        fio_args = [
            "--name=nvmept-streams",
            "--ioengine=io_uring_cmd",
            "--cmd_type=nvme",
            "--randrepeat=0",
            f"--filename={self.fio_opts['filename']}",
            f"--rw={self.fio_opts['rw']}",
            f"--output={self.filenames['output']}",
            f"--output-format={self.fio_opts['output-format']}",
        ]
        for opt in ['fixedbufs', 'nonvectored', 'force_async', 'registerfiles',
                    'sqthread_poll', 'sqthread_poll_cpu', 'hipri', 'nowait',
                    'time_based', 'runtime', 'verify', 'io_size', 'num_range',
                    'iodepth', 'iodepth_batch', 'iodepth_batch_complete',
                    'size', 'rate', 'bs', 'bssplit', 'bsrange', 'randrepeat',
                    'buffer_pattern', 'verify_pattern', 'offset', 'dataplacement',
                    'plids', 'plid_select' ]:
            if opt in self.fio_opts:
                option = f"--{opt}={self.fio_opts[opt]}"
                fio_args.append(option)

        super().setup(fio_args)


    def check_result(self):
        try:
            self._check_result()
        finally:
            release_all_streams(self.fio_opts['filename'])


    def _check_result(self):

        super().check_result()

        if 'rw' not in self.fio_opts or \
                not self.passed or \
                'json' not in self.fio_opts['output-format']:
            return

        job = self.json_data['jobs'][0]

        if self.fio_opts['rw'] in ['read', 'randread']:
            self.passed = self.check_all_ddirs(['read'], job)
        elif self.fio_opts['rw'] in ['write', 'randwrite']:
            if 'verify' not in self.fio_opts:
                self.passed = self.check_all_ddirs(['write'], job)
            else:
                self.passed = self.check_all_ddirs(['read', 'write'], job)
        elif self.fio_opts['rw'] in ['trim', 'randtrim']:
            self.passed = self.check_all_ddirs(['trim'], job)
        elif self.fio_opts['rw'] in ['readwrite', 'randrw']:
            self.passed = self.check_all_ddirs(['read', 'write'], job)
        elif self.fio_opts['rw'] in ['trimwrite', 'randtrimwrite']:
            self.passed = self.check_all_ddirs(['trim', 'write'], job)
        else:
            logging.error("Unhandled rw value %s", self.fio_opts['rw'])
            self.passed = False

        if 'iodepth' in self.fio_opts:
            # We will need to figure something out if any test uses an iodepth
            # different from 8
            if job['iodepth_level']['8'] < 95:
                logging.error("Did not achieve requested iodepth")
                self.passed = False
            else:
                logging.debug("iodepth 8 target met %s", job['iodepth_level']['8'])

        stream_ids = [int(stream) for stream in self.fio_opts['plids'].split(',')]
        if not self.check_streams(self.fio_opts['filename'], stream_ids):
            self.passed = False
            logging.error("Streams not as expected")
        else:
            logging.debug("Streams created as expected")


    def check_streams(self, dut, stream_ids):
        """
        Confirm that the specified stream IDs exist on the specified device.
        """

        id_list = get_device_stream_ids(dut)
        if not id_list:
            return False

        for stream in stream_ids:
            if stream in id_list:
                logging.debug("Stream ID %d found active on device", stream)
                id_list.remove(stream)
            else:
                if self.__class__.__name__ != "StreamsTestRand":
                    logging.error("Stream ID %d not found on device", stream)
                else:
                    logging.debug("Stream ID %d not found on device", stream)
                return False

        if len(id_list) != 0:
            logging.error("Extra stream IDs %s found on device", str(id_list))
            return False

        return True


class StreamsTestRR(StreamsTest):
    """
    NVMe pass-through test class for streams. Check to make sure output for
    selected data direction(s) is non-zero and that zero data appears for other
    directions. Check that Stream IDs are accessed in round robin order.
    """

    def check_streams(self, dut, stream_ids):
        """
        The number of IOs is less than the number of stream IDs provided. Let N
        be the number of IOs. Make sure that the device only has the first N of
        the stream IDs provided.

        This will miss some cases where some other selection algorithm happens
        to select the first N stream IDs. The solution would be to repeat this
        test multiple times. Multiple trials passing would be evidence that
        round robin is working correctly.
        """

        id_list = get_device_stream_ids(dut)
        if not id_list:
            return False

        num_streams = int(self.fio_opts['io_size'] / self.fio_opts['bs'])
        stream_ids = sorted(stream_ids)[0:num_streams]

        return super().check_streams(dut, stream_ids)


class StreamsTestRand(StreamsTest):
    """
    NVMe pass-through test class for streams. Check to make sure output for
    selected data direction(s) is non-zero and that zero data appears for other
    directions. Check that Stream IDs are accessed in random order.
    """

    def check_streams(self, dut, stream_ids):
        """
        The number of IOs is less than the number of stream IDs provided. Let N
        be the number of IOs. Confirm that the stream IDs on the device are not
        the first N stream IDs.

        This will produce false positives because it is possible for the first
        N stream IDs to be randomly selected. We can reduce the probability of
        false positives by increasing N and increasing the number of streams
        IDs to choose from, although fio has a max of 16 placement IDs.
        """

        id_list = get_device_stream_ids(dut)
        if not id_list:
            return False

        num_streams = int(self.fio_opts['io_size'] / self.fio_opts['bs'])
        stream_ids = sorted(stream_ids)[0:num_streams]

        return not super().check_streams(dut, stream_ids)


def get_device_stream_ids(dut):
    cmd = f"sudo nvme dir-receive -D 1 -O 2 -H {dut}"
    logging.debug("check streams command: %s", cmd)
    cmd = cmd.split(' ')
    cmd_result = subprocess.run(cmd, capture_output=True, check=False,
                                encoding=locale.getpreferredencoding())

    logging.debug(cmd_result.stdout)

    if cmd_result.returncode != 0:
        logging.error("Error obtaining device %s stream IDs: %s", dut, cmd_result.stderr)
        return False

    id_list = []
    for line in cmd_result.stdout.split('\n'):
        if not 'Stream Identifier' in line:
            continue
        tokens = line.split(':')
        id_list.append(int(tokens[1]))

    return id_list


def release_stream(dut, stream_id):
    """
    Release stream on given device with selected ID.
    """
    cmd = f"nvme dir-send -D 1 -O 1 -S {stream_id} {dut}"
    logging.debug("release stream command: %s", cmd)
    cmd = cmd.split(' ')
    cmd_result = subprocess.run(cmd, capture_output=True, check=False,
                                encoding=locale.getpreferredencoding())

    if cmd_result.returncode != 0:
        logging.error("Error releasing %s stream %d", dut, stream_id)
        return False

    return True


def release_all_streams(dut):
    """
    Release all streams on specified device.
    """

    id_list = get_device_stream_ids(dut)
    if not id_list:
        return False

    for stream in id_list:
        if not release_stream(dut, stream):
            return False

    return True


TEST_LIST = [
    # 4k block size
    # {seq write, rand write} x {single stream, four streams}
    {
        "test_id": 1,
        "fio_opts": {
            "rw": 'write',
            "bs": 4096,
            "io_size": 256*1024*1024,
            "verify": "crc32c",
            "plids": "8",
            "dataplacement": "streams",
            "output-format": "json",
            },
        "test_class": StreamsTest,
    },
    {
        "test_id": 2,
        "fio_opts": {
            "rw": 'randwrite',
            "bs": 4096,
            "io_size": 256*1024*1024,
            "verify": "crc32c",
            "plids": "3",
            "dataplacement": "streams",
            "output-format": "json",
            },
        "test_class": StreamsTest,
    },
    {
        "test_id": 3,
        "fio_opts": {
            "rw": 'write',
            "bs": 4096,
            "io_size": 256*1024*1024,
            "verify": "crc32c",
            "plids": "1,2,3,4",
            "dataplacement": "streams",
            "output-format": "json",
            },
        "test_class": StreamsTest,
    },
    {
        "test_id": 4,
        "fio_opts": {
            "rw": 'randwrite',
            "bs": 4096,
            "io_size": 256*1024*1024,
            "verify": "crc32c",
            "plids": "5,6,7,8",
            "dataplacement": "streams",
            "output-format": "json",
            },
        "test_class": StreamsTest,
    },
    # 256KiB block size
    # {seq write, rand write} x {single stream, four streams}
    {
        "test_id": 10,
        "fio_opts": {
            "rw": 'write',
            "bs": 256*1024,
            "io_size": 256*1024*1024,
            "verify": "crc32c",
            "plids": "88",
            "dataplacement": "streams",
            "output-format": "json",
            },
        "test_class": StreamsTest,
    },
    {
        "test_id": 11,
        "fio_opts": {
            "rw": 'randwrite',
            "bs": 256*1024,
            "io_size": 256*1024*1024,
            "verify": "crc32c",
            "plids": "20",
            "dataplacement": "streams",
            "output-format": "json",
            },
        "test_class": StreamsTest,
    },
    {
        "test_id": 12,
        "fio_opts": {
            "rw": 'write',
            "bs": 256*1024,
            "io_size": 256*1024*1024,
            "verify": "crc32c",
            "plids": "16,32,64,128",
            "dataplacement": "streams",
            "output-format": "json",
            },
        "test_class": StreamsTest,
    },
    {
        "test_id": 13,
        "fio_opts": {
            "rw": 'randwrite',
            "bs": 256*1024,
            "io_size": 256*1024*1024,
            "verify": "crc32c",
            "plids": "10,20,40,82",
            "dataplacement": "streams",
            "output-format": "json",
            },
        "test_class": StreamsTest,
    },
    # Test placement ID selection patterns
    # default is round robin
    {
        "test_id": 20,
        "fio_opts": {
            "rw": 'write',
            "bs": 4096,
            "io_size": 8192,
            "plids": '88,99,100,123,124,125,126,127,128,129,130,131,132,133,134,135',
            "dataplacement": "streams",
            "output-format": "json",
            },
        "test_class": StreamsTestRR,
    },
    {
        "test_id": 21,
        "fio_opts": {
            "rw": 'write',
            "bs": 4096,
            "io_size": 8192,
            "plids": '12,88,99,100,123,124,125,126,127,128,129,130,131,132,133,11',
            "dataplacement": "streams",
            "output-format": "json",
            },
        "test_class": StreamsTestRR,
    },
    # explicitly select round robin
    {
        "test_id": 22,
        "fio_opts": {
            "rw": 'write',
            "bs": 4096,
            "io_size": 8192,
            "plids": '22,88,99,100,123,124,125,126,127,128,129,130,131,132,133,134',
            "dataplacement": "streams",
            "output-format": "json",
            "plid_select": "roundrobin",
            },
        "test_class": StreamsTestRR,
    },
    # explicitly select random
    {
        "test_id": 23,
        "fio_opts": {
            "rw": 'write',
            "bs": 4096,
            "io_size": 8192,
            "plids": '1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16',
            "dataplacement": "streams",
            "output-format": "json",
            "plid_select": "random",
            },
        "test_class": StreamsTestRand,
    },
    # Error case with placement ID > 0xFFFF
    {
        "test_id": 30,
        "fio_opts": {
            "rw": 'write',
            "bs": 4096,
            "io_size": 8192,
            "plids": "1,2,3,0x10000",
            "dataplacement": "streams",
            "output-format": "normal",
            "plid_select": "random",
            },
        "test_class": StreamsTestRand,
        "success": SUCCESS_NONZERO,
    },
    # Error case with no stream IDs provided
    {
        "test_id": 31,
        "fio_opts": {
            "rw": 'write',
            "bs": 4096,
            "io_size": 8192,
            "dataplacement": "streams",
            "output-format": "normal",
            },
        "test_class": StreamsTestRand,
        "success": SUCCESS_NONZERO,
    },

]

def parse_args():
    """Parse command-line arguments."""

    parser = argparse.ArgumentParser()
    parser.add_argument('-d', '--debug', help='Enable debug messages', action='store_true')
    parser.add_argument('-f', '--fio', help='path to file executable (e.g., ./fio)')
    parser.add_argument('-a', '--artifact-root', help='artifact root directory')
    parser.add_argument('-s', '--skip', nargs='+', type=int,
                        help='list of test(s) to skip')
    parser.add_argument('-o', '--run-only', nargs='+', type=int,
                        help='list of test(s) to run, skipping all others')
    parser.add_argument('--dut', help='target NVMe character device to test '
                        '(e.g., /dev/ng0n1). WARNING: THIS IS A DESTRUCTIVE TEST', required=True)
    args = parser.parse_args()

    return args


def main():
    """Run tests using fio's io_uring_cmd ioengine to send NVMe pass through commands."""

    args = parse_args()

    if args.debug:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)

    artifact_root = args.artifact_root if args.artifact_root else \
        f"nvmept-streams-test-{time.strftime('%Y%m%d-%H%M%S')}"
    os.mkdir(artifact_root)
    print(f"Artifact directory is {artifact_root}")

    if args.fio:
        fio_path = str(Path(args.fio).absolute())
    else:
        fio_path = 'fio'
    print(f"fio path is {fio_path}")

    for test in TEST_LIST:
        test['fio_opts']['filename'] = args.dut

    release_all_streams(args.dut)
    test_env = {
              'fio_path': fio_path,
              'fio_root': str(Path(__file__).absolute().parent.parent),
              'artifact_root': artifact_root,
              'basename': 'nvmept-streams',
              }

    _, failed, _ = run_fio_tests(TEST_LIST, test_env, args)
    sys.exit(failed)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
#
# sgunmap-test.py
#
# Basic performance testing using fio's sg ioengine
#
# USAGE
# sgunmap-perf.py char-device block-device fio-executable
#
# EXAMPLE
# t/sgunmap-perf.py /dev/sg1 /dev/sdb ./fio
#
# REQUIREMENTS
# Python 2.6+
#
#

from __future__ import absolute_import
from __future__ import print_function
import sys
import json
import argparse
import subprocess
from six.moves import range


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('cdev',
                        help='character device target (e.g., /dev/sg0)')
    parser.add_argument('bdev',
                        help='block device target (e.g., /dev/sda)')
    parser.add_argument('fioc',
                        help='path to candidate fio executable (e.g., ./fio)')
    parser.add_argument('fior',
                        help='path to reference fio executable (e.g., ./fio)')
    args = parser.parse_args()

    return args


def fulldevice(fio, dev, ioengine='psync', rw='trim', bs='1M'):
    parameters = ["--name=test",
                  "--output-format=json",
                  "--random_generator=lfsr",
                  "--bs={0}".format(bs),
                  "--rw={0}".format(rw),
                  "--ioengine={0}".format(ioengine),
                  "--filename={0}".format(dev)]

    output = subprocess.check_output([fio] + parameters)
    jsondata = json.loads(output)
    jobdata = jsondata['jobs'][0]
    return jobdata


def runtest(fio, dev, rw, qd, batch, bs='512', runtime='30s'):
    parameters = ["--name=test",
                  "--random_generator=tausworthe64",
                  "--time_based",
                  "--runtime={0}".format(runtime),
                  "--output-format=json",
                  "--ioengine=sg",
                  "--blocksize={0}".format(bs),
                  "--rw={0}".format(rw),
                  "--filename={0}".format(dev),
                  "--iodepth={0}".format(qd),
                  "--iodepth_batch={0}".format(batch)]

    output = subprocess.check_output([fio] + parameters)
    jsondata = json.loads(output)
    jobdata = jsondata['jobs'][0]
#    print(parameters)

    return jobdata


def runtests(fio, dev, qd, batch, rw, bs='512', trials=5):
    iops = []
    for x in range(trials):
        jd = runtest(fio, dev, rw, qd, batch, bs=bs)
        total = jd['read']['iops'] + jd['write']['iops'] + jd['trim']['iops']
#       print(total)
        iops.extend([total])
    return iops, (sum(iops) / trials)

if __name__ == '__main__':
    args = parse_args()

    print("Trimming full device {0}".format(args.cdev))
    fulldevice(args.fior, args.cdev, ioengine='sg')

    print("Running rand read tests on {0}"
        " with fio candidate build {1}".format(args.cdev, args.fioc))
    randread, rrmean = runtests(args.fioc, args.cdev, 16, 1, 'randread',
        trials=5)
    print("IOPS mean {0}, trials {1}".format(rrmean, randread))

    print("Running rand read tests on {0}"
        " with fio reference build {1}".format(args.cdev, args.fior))
    randread, rrmean = runtests(args.fior, args.cdev, 16, 1, 'randread',
        trials=5)
    print("IOPS mean {0}, trials {1}".format(rrmean, randread))

    print("Running rand write tests on {0}"
        " with fio candidate build {1}".format(args.cdev, args.fioc))
    randwrite, rwmean = runtests(args.fioc, args.cdev, 16, 1, 'randwrite',
        trials=5)
    print("IOPS mean {0}, trials {1}".format(rwmean, randwrite))

    print("Running rand write tests on {0}"
        " with fio reference build {1}".format(args.cdev, args.fior))
    randwrite, rwmean = runtests(args.fior, args.cdev, 16, 1, 'randwrite',
        trials=5)
    print("IOPS mean {0}, trials {1}".format(rwmean, randwrite))



#!/usr/bin/env python3

"""
# strided.py
#
# Test zonemode=strided. This uses the null ioengine when no file is
# specified. If a file is specified, use it for randdom read testing.
# Some of the zoneranges in the tests are 16MiB. So when using a file
# a minimum size of 64MiB is recommended.
#
# USAGE
# python strided.py fio-executable [-f file/device]
#
# EXAMPLES
# python t/strided.py ./fio
# python t/strided.py ./fio -f /dev/sda
# dd if=/dev/zero of=temp bs=1M count=64
# python t/strided.py ./fio -f temp
#
# ===TEST MATRIX===
#
# --zonemode=strided, zoneskip unset
#   w/ randommap and LFSR
#       zonesize=zonerange  all blocks in zonerange touched
#       zonesize>zonerange  all blocks touched and roll-over back into zone
#       zonesize<zonerange  all blocks inside zone
#
#   w/o randommap       all blocks inside zone
"""

import os
import sys
import time
import argparse
from pathlib import Path
from fiotestlib import FioJobCmdTest, run_fio_tests


class StridedTest(FioJobCmdTest):
    """Test zonemode=strided."""

    def setup(self, parameters):
        fio_args = [
                    "--name=strided",
                    "--zonemode=strided",
                    "--log_offset=1",
                    "--randrepeat=0",
                    "--rw=randread",
                    f"--write_iops_log={self.filenames['iopslog']}",
                    f"--output={self.filenames['output']}",
                    f"--zonerange={self.fio_opts['zonerange']}",
                    f"--zonesize={self.fio_opts['zonesize']}",
                    f"--bs={self.fio_opts['bs']}",
                   ]

        for opt in ['norandommap', 'random_generator', 'offset']:
            if opt in self.fio_opts:
                option = f"--{opt}={self.fio_opts[opt]}"
                fio_args.append(option)

        if 'filename' in self.fio_opts:
            for opt in ['filename', 'filesize']:
                option = f"--{opt}={self.fio_opts[opt]}"
                fio_args.append(option)
        else:
            fio_args.append('--ioengine=null')
            for opt in ['size', 'io_size', 'filesize']:
                option = f"--{opt}={self.fio_opts[opt]}"
                fio_args.append(option)

        super().setup(fio_args)

    def check_result(self):
        super().check_result()
        if not self.passed:
            return

        zonestart = 0 if 'offset' not in self.fio_opts else self.fio_opts['offset']
        iospersize = self.fio_opts['zonesize'] / self.fio_opts['bs']
        iosperrange = self.fio_opts['zonerange'] / self.fio_opts['bs']
        iosperzone = 0
        lines = self.iops_log_lines.split('\n')
        zoneset = set()

        for line in lines:
            if len(line) == 0:
                continue

            if iosperzone == iospersize:
                # time to move to a new zone
                iosperzone = 0
                zoneset = set()
                zonestart += self.fio_opts['zonerange']
                if zonestart >= self.fio_opts['filesize']:
                    zonestart = 0 if 'offset' not in self.fio_opts else self.fio_opts['offset']

            iosperzone = iosperzone + 1
            tokens = line.split(',')
            offset = int(tokens[4])
            if offset < zonestart or offset >= zonestart + self.fio_opts['zonerange']:
                print(f"Offset {offset} outside of zone starting at {zonestart}")
                return

            # skip next section if norandommap is enabled with no
            # random_generator or with a random_generator != lfsr
            if 'norandommap' in self.fio_opts:
                if 'random_generator' in self.fio_opts:
                    if self.fio_opts['random_generator'] != 'lfsr':
                        continue
                else:
                    continue

            # we either have a random map enabled or we
            # are using an LFSR
            # so all blocks should be unique and we should have
            # covered the entire zone when iosperzone % iosperrange == 0
            block = (offset - zonestart) / self.fio_opts['bs']
            if block in zoneset:
                print(f"Offset {offset} in zone already touched")
                return

            zoneset.add(block)
            if iosperzone % iosperrange == 0:
                if len(zoneset) != iosperrange:
                    print(f"Expected {iosperrange} blocks in zone but only saw {len(zoneset)}")
                    return
                zoneset = set()


TEST_LIST = [   # randommap enabled
    {
        "test_id": 1,
        "fio_opts": {
            "zonerange": 4096,
            "zonesize": 4096,
            "bs": 4096,
            "offset": 8*4096,
            "size": 16*4096,
            "io_size": 16*4096,
            },
        "test_class": StridedTest,
    },
    {
        "test_id": 2,
        "fio_opts": {
            "zonerange": 4096,
            "zonesize": 4096,
            "bs": 4096,
            "size": 16*4096,
            "io_size": 16*4096,
            },
        "test_class": StridedTest,
    },
    {
        "test_id": 3,
        "fio_opts": {
            "zonerange": 16*1024*1024,
            "zonesize": 16*1024*1024,
            "bs": 4096,
            "size": 256*1024*1024,
            "io_size": 256*1024*204,
            },
        "test_class": StridedTest,
    },
    {
        "test_id": 4,
        "fio_opts": {
            "zonerange": 4096,
            "zonesize": 4*4096,
            "bs": 4096,
            "size": 16*4096,
            "io_size": 16*4096,
            },
        "test_class": StridedTest,
    },
    {
        "test_id": 5,
        "fio_opts": {
            "zonerange": 16*1024*1024,
            "zonesize": 32*1024*1024,
            "bs": 4096,
            "size": 256*1024*1024,
            "io_size": 256*1024*204,
            },
        "test_class": StridedTest,
    },
    {
        "test_id": 6,
        "fio_opts": {
            "zonerange": 8192,
            "zonesize": 4096,
            "bs": 4096,
            "size": 16*4096,
            "io_size": 16*4096,
            },
        "test_class": StridedTest,
    },
    {
        "test_id": 7,
        "fio_opts": {
            "zonerange": 16*1024*1024,
            "zonesize": 8*1024*1024,
            "bs": 4096,
            "size": 256*1024*1024,
            "io_size": 256*1024*204,
            },
        "test_class": StridedTest,
    },
            # lfsr
    {
        "test_id": 8,
        "fio_opts": {
            "random_generator": "lfsr",
            "zonerange": 4096*1024,
            "zonesize": 4096*1024,
            "bs": 4096,
            "offset": 8*4096*1024,
            "size": 16*4096*1024,
            "io_size": 16*4096*1024,
            },
        "test_class": StridedTest,
    },
    {
        "test_id": 9,
        "fio_opts": {
            "random_generator": "lfsr",
            "zonerange": 4096*1024,
            "zonesize": 4096*1024,
            "bs": 4096,
            "size": 16*4096*1024,
            "io_size": 16*4096*1024,
            },
        "test_class": StridedTest,
    },
    {
        "test_id": 10,
        "fio_opts": {
            "random_generator": "lfsr",
            "zonerange": 16*1024*1024,
            "zonesize": 16*1024*1024,
            "bs": 4096,
            "size": 256*1024*1024,
            "io_size": 256*1024*204,
            },
        "test_class": StridedTest,
    },
    {
        "test_id": 11,
        "fio_opts": {
            "random_generator": "lfsr",
            "zonerange": 4096*1024,
            "zonesize": 4*4096*1024,
            "bs": 4096,
            "size": 16*4096*1024,
            "io_size": 16*4096*1024,
            },
        "test_class": StridedTest,
    },
    {
        "test_id": 12,
        "fio_opts": {
            "random_generator": "lfsr",
            "zonerange": 16*1024*1024,
            "zonesize": 32*1024*1024,
            "bs": 4096,
            "size": 256*1024*1024,
            "io_size": 256*1024*204,
            },
        "test_class": StridedTest,
    },
    {
        "test_id": 13,
        "fio_opts": {
            "random_generator": "lfsr",
            "zonerange": 8192*1024,
            "zonesize": 4096*1024,
            "bs": 4096,
            "size": 16*4096*1024,
            "io_size": 16*4096*1024,
            },
        "test_class": StridedTest,
    },
    {
        "test_id": 14,
        "fio_opts": {
            "random_generator": "lfsr",
            "zonerange": 16*1024*1024,
            "zonesize": 8*1024*1024,
            "bs": 4096,
            "size": 256*1024*1024,
            "io_size": 256*1024*204,
            },
        "test_class": StridedTest,
    },
    # norandommap
    {
        "test_id": 15,
        "fio_opts": {
            "norandommap": 1,
            "zonerange": 4096,
            "zonesize": 4096,
            "bs": 4096,
            "offset": 8*4096,
            "size": 16*4096,
            "io_size": 16*4096,
            },
        "test_class": StridedTest,
    },
    {
        "test_id": 16,
        "fio_opts": {
            "norandommap": 1,
            "zonerange": 4096,
            "zonesize": 4096,
            "bs": 4096,
            "size": 16*4096,
            "io_size": 16*4096,
            },
        "test_class": StridedTest,
    },
    {
        "test_id": 17,
        "fio_opts": {
            "norandommap": 1,
            "zonerange": 16*1024*1024,
            "zonesize": 16*1024*1024,
            "bs": 4096,
            "size": 256*1024*1024,
            "io_size": 256*1024*204,
            },
        "test_class": StridedTest,
    },
    {
        "test_id": 18,
        "fio_opts": {
            "norandommap": 1,
            "zonerange": 4096,
            "zonesize": 8192,
            "bs": 4096,
            "size": 16*4096,
            "io_size": 16*4096,
            },
        "test_class": StridedTest,
    },
    {
        "test_id": 19,
        "fio_opts": {
            "norandommap": 1,
            "zonerange": 16*1024*1024,
            "zonesize": 32*1024*1024,
            "bs": 4096,
            "size": 256*1024*1024,
            "io_size": 256*1024*204,
            },
        "test_class": StridedTest,
    },
    {
        "test_id": 20,
        "fio_opts": {
            "norandommap": 1,
            "zonerange": 8192,
            "zonesize": 4096,
            "bs": 4096,
            "size": 16*4096,
            "io_size": 16*4096,
            },
        "test_class": StridedTest,
    },
    {
        "test_id": 21,
        "fio_opts": {
            "norandommap": 1,
            "zonerange": 16*1024*1024,
            "zonesize": 8*1024*1024,
            "bs": 4096,
            "size": 256*1024*1024,
            "io_size": 256*1024*1024,
            },
        "test_class": StridedTest,
    },
]


def parse_args():
    """Parse command-line arguments."""

    parser = argparse.ArgumentParser()
    parser.add_argument('-f', '--fio', help='path to file executable (e.g., ./fio)')
    parser.add_argument('-a', '--artifact-root', help='artifact root directory')
    parser.add_argument('-s', '--skip', nargs='+', type=int,
                        help='list of test(s) to skip')
    parser.add_argument('-o', '--run-only', nargs='+', type=int,
                        help='list of test(s) to run, skipping all others')
    parser.add_argument('--dut',
                        help='target file/device to test.')
    args = parser.parse_args()

    return args


def main():
    """Run zonemode=strided tests."""

    args = parse_args()

    artifact_root = args.artifact_root if args.artifact_root else \
        f"strided-test-{time.strftime('%Y%m%d-%H%M%S')}"
    os.mkdir(artifact_root)
    print(f"Artifact directory is {artifact_root}")

    if args.fio:
        fio_path = str(Path(args.fio).absolute())
    else:
        fio_path = 'fio'
    print(f"fio path is {fio_path}")

    if args.dut:
        statinfo = os.stat(args.dut)
        filesize = statinfo.st_size
        if filesize == 0:
            f = os.open(args.dut, os.O_RDONLY)
            filesize = os.lseek(f, 0, os.SEEK_END)
            os.close(f)

    for test in TEST_LIST:
        if args.dut:
            test['fio_opts']['filename'] = os.path.abspath(args.dut)
            test['fio_opts']['filesize'] = filesize
        else:
            test['fio_opts']['filesize'] = test['fio_opts']['size']

    test_env = {
              'fio_path': fio_path,
              'fio_root': str(Path(__file__).absolute().parent.parent),
              'artifact_root': artifact_root,
              'basename': 'strided',
              }

    _, failed, _ = run_fio_tests(TEST_LIST, test_env, args)
    sys.exit(failed)


if __name__ == '__main__':
    main()
